from __future__ import annotations

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from django_spire.history.choices import HistoryEventChoices
from django_spire.history.models import HistoryEvent
from django_spire.history.querysets import HistoryQuerySet


class TestHistoryEventChoices(TestCase):
    def test_active_choice(self) -> None:
        assert HistoryEventChoices.ACTIVE.value == 'acti'
        assert HistoryEventChoices.ACTIVE.label == 'Active'

    def test_all_choices_exist(self) -> None:
        choices = [c[0] for c in HistoryEventChoices.choices]

        assert 'crea' in choices
        assert 'upda' in choices
        assert 'acti' in choices
        assert 'inac' in choices
        assert 'dele' in choices
        assert 'unde' in choices

    def test_created_choice(self) -> None:
        assert HistoryEventChoices.CREATED.value == 'crea'
        assert HistoryEventChoices.CREATED.label == 'Created'

    def test_deleted_choice(self) -> None:
        assert HistoryEventChoices.DELETED.value == 'dele'
        assert HistoryEventChoices.DELETED.label == 'Deleted'

    def test_inactive_choice(self) -> None:
        assert HistoryEventChoices.INACTIVE.value == 'inac'
        assert HistoryEventChoices.INACTIVE.label == 'Inactive'

    def test_undeleted_choice(self) -> None:
        assert HistoryEventChoices.UNDELETED.value == 'unde'
        assert HistoryEventChoices.UNDELETED.label == 'Un-Deleted'

    def test_updated_choice(self) -> None:
        assert HistoryEventChoices.UPDATED.value == 'upda'
        assert HistoryEventChoices.UPDATED.label == 'Updated'


class TestHistoryEvent(TestCase):
    def setUp(self) -> None:
        self.user = User.objects.create_user(
            username='testuser',
            password='testpass'  # noqa: S106
        )
        self.content_type = ContentType.objects.get_for_model(User)

    def test_create_history_event(self) -> None:
        event = HistoryEvent.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            event=HistoryEventChoices.CREATED
        )

        assert event.pk is not None
        assert event.event == 'crea'
        assert event.content_object == self.user

    def test_event_verbose_property(self) -> None:
        event = HistoryEvent.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            event=HistoryEventChoices.UPDATED
        )

        assert event.event_verbose == 'Updated'

    def test_str_representation(self) -> None:
        event = HistoryEvent.objects.create(
            content_type=self.content_type,
            object_id=self.user.pk,
            event=HistoryEventChoices.CREATED
        )

        assert 'Created' in str(event)


class TestHistoryQuerySet(TestCase):
    def test_active_filter(self) -> None:
        # HistoryQuerySet is meant to be used with models that have
        # is_active and is_deleted fields (via HistoryModelMixin)
        qs = HistoryQuerySet(model=HistoryEvent)

        # Verify the method exists and returns a QuerySet
        assert hasattr(qs, 'active')

    def test_deleted_filter(self) -> None:
        qs = HistoryQuerySet(model=HistoryEvent)
        assert hasattr(qs, 'deleted')

    def test_inactive_filter(self) -> None:
        qs = HistoryQuerySet(model=HistoryEvent)
        assert hasattr(qs, 'inactive')



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.choices import (
    NotificationPriorityChoices,
    NotificationStatusChoices,
    NotificationTypeChoices,
)
from django_spire.notification.tests.factories import create_test_notification


class NotificationModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.notification = create_test_notification()

    def test_default_status_is_pending(self):
        assert self.notification.status == NotificationStatusChoices.PENDING

    def test_default_priority_is_low(self):
        assert self.notification.priority == NotificationPriorityChoices.LOW

    def test_get_type_display(self):
        self.notification.type = NotificationTypeChoices.APP
        assert self.notification.get_type_display() == 'App'

        self.notification.type = NotificationTypeChoices.EMAIL
        assert self.notification.get_type_display() == 'Email'

        self.notification.type = NotificationTypeChoices.SMS
        assert self.notification.get_type_display() == 'Sms'

        self.notification.type = NotificationTypeChoices.PUSH
        assert self.notification.get_type_display() == 'Push'

    def test_get_status_display(self):
        self.notification.status = NotificationStatusChoices.PENDING
        assert self.notification.get_status_display() == 'Pending'

        self.notification.status = NotificationStatusChoices.PROCESSING
        assert self.notification.get_status_display() == 'Processing'

        self.notification.status = NotificationStatusChoices.SENT
        assert self.notification.get_status_display() == 'Sent'

        self.notification.status = NotificationStatusChoices.ERRORED
        assert self.notification.get_status_display() == 'Errored'

        self.notification.status = NotificationStatusChoices.FAILED
        assert self.notification.get_status_display() == 'Failed'

    def test_get_priority_display(self):
        self.notification.priority = NotificationPriorityChoices.LOW
        assert self.notification.get_priority_display() == 'Low'

        self.notification.priority = NotificationPriorityChoices.MEDIUM
        assert self.notification.get_priority_display() == 'Medium'

        self.notification.priority = NotificationPriorityChoices.HIGH
        assert self.notification.get_priority_display() == 'High'

    def test_user_relationship(self):
        assert self.notification.user is not None
        assert self.notification.user.pk is not None

    def test_created_datetime_auto_set(self):
        assert self.notification.created_datetime is not None

    def test_sent_datetime_null_by_default(self):
        assert self.notification.sent_datetime is None

    def test_content_object_null_by_default(self):
        assert self.notification.content_object is None
        assert self.notification.content_type is None
        assert self.notification.object_id is None



from __future__ import annotations

from unittest.mock import MagicMock

from django.contrib.sessions.middleware import SessionMiddleware
from django.db.models import QuerySet
from django.test import RequestFactory, TestCase

from django_spire.contrib.queryset.enums import SessionFilterActionEnum
from django_spire.contrib.queryset.filter_tools import filter_by_lookup_map
from django_spire.contrib.queryset.mixins import SearchQuerySetMixin, SessionFilterQuerySetMixin


class TestSessionFilterActionEnum(TestCase):
    def test_clear_value(self) -> None:
        assert SessionFilterActionEnum.CLEAR == 'Clear'

    def test_filter_value(self) -> None:
        assert SessionFilterActionEnum.FILTER == 'Filter'

    def test_is_str_enum(self) -> None:
        assert isinstance(SessionFilterActionEnum.CLEAR, str)
        assert isinstance(SessionFilterActionEnum.FILTER, str)


class TestFilterByLookupMap(TestCase):
    def setUp(self) -> None:
        self.queryset = MagicMock(spec=QuerySet)
        self.queryset.filter.return_value = self.queryset

    def test_applies_filter_from_lookup_map(self) -> None:
        lookup_map = {'name': 'name__icontains'}
        data = {'name': 'test'}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(name__icontains='test')

    def test_extra_filters_applied(self) -> None:
        lookup_map = {'name': 'name__icontains'}
        data = {'name': 'test'}
        extra_filters = {'is_active': True}

        filter_by_lookup_map(self.queryset, lookup_map, data, extra_filters)

        self.queryset.filter.assert_called_once_with(
            name__icontains='test',
            is_active=True
        )

    def test_extra_filters_defaults_to_empty_dict(self) -> None:
        lookup_map = {}
        data = {}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with()

    def test_ignores_empty_string_values(self) -> None:
        lookup_map = {'name': 'name__icontains', 'email': 'email__icontains'}
        data = {'name': 'test', 'email': ''}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(name__icontains='test')

    def test_ignores_keys_not_in_lookup_map(self) -> None:
        lookup_map = {'name': 'name__icontains'}
        data = {'name': 'test', 'other': 'value'}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(name__icontains='test')

    def test_ignores_none_values(self) -> None:
        lookup_map = {'name': 'name__icontains', 'email': 'email__icontains'}
        data = {'name': 'test', 'email': None}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(name__icontains='test')

    def test_ignores_empty_list_values(self) -> None:
        lookup_map = {'name': 'name__icontains', 'tags': 'tags__in'}
        data = {'name': 'test', 'tags': []}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(name__icontains='test')

    def test_multiple_filters_applied(self) -> None:
        lookup_map = {'name': 'name__icontains', 'status': 'status'}
        data = {'name': 'test', 'status': 'active'}

        filter_by_lookup_map(self.queryset, lookup_map, data)

        self.queryset.filter.assert_called_once_with(
            name__icontains='test',
            status='active'
        )

    def test_returns_filtered_queryset(self) -> None:
        lookup_map = {'name': 'name__icontains'}
        data = {'name': 'test'}

        result = filter_by_lookup_map(self.queryset, lookup_map, data)

        assert result == self.queryset


class TestSessionFilterQuerySetMixin(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()

    def _create_request_with_session(self, path: str = '/', data: dict | None = None):
        request = self.factory.get(path, data or {})
        middleware = SessionMiddleware(lambda req: None)
        middleware.process_request(request)
        request.session.save()
        return request

    def test_has_bulk_filter_abstract_method(self) -> None:
        assert hasattr(SessionFilterQuerySetMixin, 'bulk_filter')

    def test_has_process_session_filter_method(self) -> None:
        assert hasattr(SessionFilterQuerySetMixin, 'process_session_filter')

    def test_is_queryset_subclass(self) -> None:
        assert issubclass(SessionFilterQuerySetMixin, QuerySet)


class TestSearchQuerySetMixin(TestCase):
    def test_has_search_abstract_method(self) -> None:
        assert hasattr(SearchQuerySetMixin, 'search')

    def test_is_queryset_subclass(self) -> None:
        assert issubclass(SearchQuerySetMixin, QuerySet)



from __future__ import annotations

from django.urls import reverse

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase


class GroupPageViewsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')

    def test_list_view_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        assert response.status_code == 403

    def test_list_view_with_permission(self) -> None:
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        assert response.status_code == 200

    def test_list_view_context_contains_groups(self) -> None:
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        assert 'group_list' in response.context

    def test_list_view_contains_created_group(self) -> None:
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        group_names = [g.name for g in response.context['group_list']]
        assert 'Test Group' in group_names

    def test_detail_view_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 403

    def test_detail_view_with_permission(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 200

    def test_detail_view_context_contains_group(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert 'group' in response.context
        assert response.context['group'] == self.group

    def test_detail_view_404_for_nonexistent_group(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': 99999})
        )
        assert response.status_code == 404

    def test_detail_view_contains_permission_data(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert 'permission_data' in response.context

    def test_detail_view_contains_user_lists(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert 'active_user_list' in response.context
        assert 'inactive_user_list' in response.context

    def test_list_view_contains_permission_data(self) -> None:
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        assert 'group_list_permission_data' in response.context

    def test_list_view_multiple_groups(self) -> None:
        AuthGroup.objects.create(name='Group 2')
        AuthGroup.objects.create(name='Group 3')
        response = self.client.get(reverse('django_spire:auth:group:page:list'))
        assert len(response.context['group_list']) >= 3

    def test_detail_view_with_users_in_group(self) -> None:
        user = create_user(username='groupmember')
        self.group.user_set.add(user)
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 200

    def test_detail_view_separates_active_inactive_users(self) -> None:
        active_user = create_user(username='activeuser', is_active=True)
        inactive_user = create_user(username='inactiveuser', is_active=False)
        self.group.user_set.add(active_user, inactive_user)
        response = self.client.get(
            reverse('django_spire:auth:group:page:detail', kwargs={'pk': self.group.pk})
        )
        active_ids = [u.pk for u in response.context['active_user_list']]
        inactive_ids = [u.pk for u in response.context['inactive_user_list']]
        assert active_user.pk in active_ids
        assert inactive_user.pk in inactive_ids


class GroupFormViewsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')

    def test_add_form_view_get(self) -> None:
        response = self.client.get(reverse('django_spire:auth:group:form:add'))
        assert response.status_code == 200

    def test_add_form_view_post_creates_group(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': 'New Group'}
        )
        assert response.status_code == 302
        assert AuthGroup.objects.filter(name='New Group').exists()

    def test_add_form_view_post_invalid_name(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': 'All Users'}
        )
        assert response.status_code == 200
        assert not AuthGroup.objects.filter(name='All Users').exists()

    def test_add_form_view_post_empty_name(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': ''}
        )
        assert response.status_code == 200

    def test_update_form_view_get(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:update', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 200

    def test_update_form_view_post_updates_group(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:update', kwargs={'pk': self.group.pk}),
            data={'name': 'Updated Group'}
        )
        assert response.status_code == 302
        self.group.refresh_from_db()
        assert self.group.name == 'Updated Group'

    def test_update_form_view_nonexistent_group_returns_200(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:update', kwargs={'pk': 99999})
        )
        assert response.status_code == 200

    def test_delete_form_view_get(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:delete', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 200

    def test_delete_form_view_404(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:delete', kwargs={'pk': 99999})
        )
        assert response.status_code == 404

    def test_user_form_view_get(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:user', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 200

    def test_user_form_view_404(self) -> None:
        response = self.client.get(
            reverse('django_spire:auth:group:form:user', kwargs={'pk': 99999})
        )
        assert response.status_code == 404

    def test_add_form_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(reverse('django_spire:auth:group:form:add'))
        assert response.status_code == 403

    def test_update_form_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(
            reverse('django_spire:auth:group:form:update', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 403

    def test_delete_form_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(
            reverse('django_spire:auth:group:form:delete', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 403

    def test_user_form_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.get(
            reverse('django_spire:auth:group:form:user', kwargs={'pk': self.group.pk})
        )
        assert response.status_code == 403

    def test_add_form_view_post_duplicate_name(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': 'Test Group'}
        )
        assert response.status_code == 200

    def test_update_form_view_post_to_reserved_name(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:update', kwargs={'pk': self.group.pk}),
            data={'name': 'All Users'}
        )
        assert response.status_code == 200
        self.group.refresh_from_db()
        assert self.group.name != 'All Users'

    def test_add_form_view_post_unicode_name(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': 'Tëst Grøup 日本語'}
        )
        assert response.status_code == 302
        assert AuthGroup.objects.filter(name='Tëst Grøup 日本語').exists()

    def test_add_form_view_post_special_characters(self) -> None:
        response = self.client.post(
            reverse('django_spire:auth:group:form:add'),
            data={'name': 'Group & Co <Test>'}
        )
        assert response.status_code == 302

    def test_user_form_view_post_adds_users(self) -> None:
        user = create_user(username='newgroupuser')
        response = self.client.post(
            reverse('django_spire:auth:group:form:user', kwargs={'pk': self.group.pk}),
            data={'users': [user.pk]}
        )
        assert response.status_code == 302
        assert user in self.group.user_set.all()

    def test_user_form_view_post_removes_users(self) -> None:
        user = create_user(username='removeuser')
        self.group.user_set.add(user)
        response = self.client.post(
            reverse('django_spire:auth:group:form:user', kwargs={'pk': self.group.pk}),
            data={'users': []}
        )
        assert response.status_code == 302
        assert user not in self.group.user_set.all()


class GroupJsonViewsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')

    def test_permission_form_ajax_invalid_app(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'invalid_app'}
            ),
            data={'perm_level': 'View'},
            content_type='application/json'
        )
        assert response.status_code == 200
        data = response.json()
        assert data['status'] == 400

    def test_permission_form_ajax_valid_app(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'group'}
            ),
            data='{"perm_level": "View"}',
            content_type='application/json'
        )
        assert response.status_code == 200

    def test_permission_form_ajax_requires_post(self) -> None:
        response = self.client.get(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'group'}
            )
        )
        assert response.status_code == 405

    def test_permission_form_ajax_404_group(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': 99999, 'app_name': 'group'}
            ),
            data='{"perm_level": "View"}',
            content_type='application/json'
        )
        assert response.status_code == 404

    def test_special_role_ajax_invalid_app(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_special_role_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'invalid_app'}
            ),
            data='{"codename": "test", "grant_special_role_access": true}',
            content_type='application/json'
        )
        assert response.status_code == 200
        data = response.json()
        assert data['status'] == 400

    def test_permission_form_ajax_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'group'}
            ),
            data='{"perm_level": "View"}',
            content_type='application/json'
        )
        assert response.status_code == 403

    def test_special_role_ajax_requires_permission(self) -> None:
        user = create_user(username='normaluser')
        self.client.force_login(user)
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_special_role_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'group'}
            ),
            data='{"codename": "test", "grant_special_role_access": true}',
            content_type='application/json'
        )
        assert response.status_code == 403

    def test_permission_form_ajax_all_perm_levels(self) -> None:
        for perm_level in ['None', 'View', 'Add', 'Change', 'Delete']:
            response = self.client.post(
                reverse(
                    'django_spire:auth:group:json:group_permission_ajax',
                    kwargs={'pk': self.group.pk, 'app_name': 'group'}
                ),
                data=f'{{"perm_level": "{perm_level}"}}',
                content_type='application/json'
            )
            assert response.status_code == 200

    def test_permission_form_ajax_invalid_perm_level_defaults_to_none(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:auth:group:json:group_permission_ajax',
                kwargs={'pk': self.group.pk, 'app_name': 'group'}
            ),
            data='{"perm_level": "InvalidLevel"}',
            content_type='application/json'
        )
        assert response.status_code == 200
        data = response.json()
        assert data['status'] == 200



from __future__ import annotations

import logging
import time

from typing import Callable, ParamSpec, TypeVar

from django.conf import settings


P = ParamSpec('P')
R = TypeVar('R')

log = logging.getLogger(__name__)


def performance_timer(func: Callable[P, R]) -> Callable[P, R]:
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        if settings.DEBUG:
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            end_time = time.perf_counter()

            message = f'{end_time - start_time:.4f} seconds runtime for "{func.__module__}.{func.__qualname__}"'
            log.warning(message)

            return result

        return func(*args, **kwargs)

    return wrapper



from __future__ import annotations

from enum import Enum
from unittest import TestCase

from django_spire.core.converters.to_enums import django_choices_to_enums


class TestDjangoChoicesToEnums(TestCase):
    def test_basic_conversion(self) -> None:
        """Test basic conversion with simple choices"""

        choices = [
            ('ACTIVE', 'Active'),
            ('INACTIVE', 'Inactive'),
            ('PENDING', 'Pending')
        ]

        ResultEnum = django_choices_to_enums('Status', choices)

        assert issubclass(ResultEnum, Enum)
        assert ResultEnum.__name__ == 'Status'

        assert ResultEnum.ACTIVE.value == 'ACTIVE'
        assert ResultEnum.INACTIVE.value == 'INACTIVE'
        assert ResultEnum.PENDING.value == 'PENDING'

        assert len(ResultEnum) == 3

    def test_different_key_value(self) -> None:
        """Test when keys and values differ"""

        choices = [
            ('A', 'Alpha'),
            ('B', 'Beta'),
            ('G', 'Gamma')
        ]

        ResultEnum = django_choices_to_enums('Greek', choices)

        assert issubclass(ResultEnum, Enum)
        assert ResultEnum.A.value == 'A'
        assert ResultEnum.B.value == 'B'
        assert ResultEnum.G.value == 'G'
        assert len(ResultEnum) == 3

    def test_duplicate_keys(self) -> None:
        """Test behavior with duplicate keys (should use last occurrence)"""

        choices = [
            ('STATUS', 'First'),
            ('STATUS', 'Second')
        ]

        ResultEnum = django_choices_to_enums('DuplicateStatus', choices)

        assert issubclass(ResultEnum, Enum)
        assert ResultEnum.STATUS.value == 'STATUS'
        assert len(ResultEnum) == 1

    def test_empty_choices(self) -> None:
        """Test function with empty choices list"""

        choices = []
        ResultEnum = django_choices_to_enums('EmptyStatus', choices)

        assert issubclass(ResultEnum, Enum)
        assert ResultEnum.__name__ == 'EmptyStatus'
        assert len(ResultEnum) == 0

    def test_lowercase_keys_converted_to_uppercase(self) -> None:
        """Test that lowercase keys are converted to uppercase"""

        choices = [
            ('active', 'Active'),
            ('inactive', 'Inactive')
        ]

        ResultEnum = django_choices_to_enums('Status', choices)

        assert hasattr(ResultEnum, 'ACTIVE')
        assert hasattr(ResultEnum, 'INACTIVE')
        assert ResultEnum.ACTIVE.value == 'active'
        assert ResultEnum.INACTIVE.value == 'inactive'

    def test_single_choice(self) -> None:
        """Test with a single choice"""

        choices = [('OPEN', 'Open')]
        ResultEnum = django_choices_to_enums('SingleStatus', choices)

        assert issubclass(ResultEnum, Enum)
        assert ResultEnum.__name__ == 'SingleStatus'
        assert ResultEnum.OPEN.value == 'OPEN'
        assert len(ResultEnum) == 1



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.collection.tests.factories import create_test_collection
from django_spire.knowledge.entry.models import Entry


class EntryFactoryServiceTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.collection = create_test_collection()

    def test_create_from_files_empty_list(self):
        entries = Entry.services.factory.create_from_files(
            author=self.super_user,
            collection=self.collection,
            files=[]
        )
        assert entries == []



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.file.fields import MultipleFileField
from django_spire.file.forms import FileForm


class FileFormTests(BaseTestCase):
    def test_has_files_field(self):
        assert 'files' in FileForm().fields

    def test_files_field_is_multiple_file_field(self):
        form = FileForm()

        assert isinstance(form.fields['files'], MultipleFileField)

    def test_form_field_count(self):
        form = FileForm()

        assert len(form.fields) == 1



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.choices import EntryVersionStatusChoices
from django_spire.knowledge.entry.version.block.models import EntryVersionBlock
from django_spire.knowledge.entry.version.block.tests.factories import (
    create_test_block_form_data,
    create_test_version_block,
)
from django_spire.knowledge.entry.version.tests.factories import create_test_entry_version


class EntryVersionProcessorServiceTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.entry_version = create_test_entry_version()

    def test_publish(self):
        self.entry_version.services.processor.publish()
        self.entry_version.refresh_from_db()
        assert self.entry_version.status == EntryVersionStatusChoices.PUBLISHED
        assert self.entry_version.published_datetime is not None

    def test_add_update_delete_blocks_add_new(self):
        block_data = create_test_block_form_data(id='new_block_123')
        self.entry_version.services.processor.add_update_delete_blocks([block_data])
        assert EntryVersionBlock.objects.filter(version=self.entry_version).count() == 1

    def test_add_update_delete_blocks_update_existing(self):
        existing_block = create_test_version_block(version=self.entry_version)
        block_data = create_test_block_form_data(
            id=existing_block.id,
            data={'text': 'updated text'}
        )
        self.entry_version.services.processor.add_update_delete_blocks([block_data])

        existing_block.refresh_from_db()
        assert existing_block._block_data['text'] == 'updated text'

    def test_add_update_delete_blocks_delete_missing(self):
        existing_block = create_test_version_block(version=self.entry_version)
        self.entry_version.services.processor.add_update_delete_blocks([])
        assert not EntryVersionBlock.objects.filter(pk=existing_block.pk).exists()

    def test_add_update_delete_blocks_mixed_operations(self):
        existing_block = create_test_version_block(version=self.entry_version)
        block_to_delete = create_test_version_block(version=self.entry_version, order=2)

        block_data_update = create_test_block_form_data(
            id=existing_block.id,
            order=0,
            data={'text': 'updated'}
        )
        block_data_new = create_test_block_form_data(id='new_block', order=1)

        self.entry_version.services.processor.add_update_delete_blocks([
            block_data_update,
            block_data_new
        ])

        assert EntryVersionBlock.objects.filter(version=self.entry_version).count() == 2
        assert not EntryVersionBlock.objects.filter(pk=block_to_delete.pk).exists()



from __future__ import annotations

import json

from django.forms import Widget
from django.template import loader
from django.utils.safestring import SafeString, mark_safe


class MultipleWidget(Widget):
    needs_multipart_form = True
    template_name = 'django_spire/file/widget/multiple_file_widget.html'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def render(self, name: str, value, attrs: dict | None = None, renderer=None) -> SafeString:
        context = self.get_context(name, value, attrs)
        template = loader.get_template(self.template_name).render(context)
        return mark_safe(template)

    def value_from_datadict(self, data: dict, files, name: str) -> list[dict]:
        return json.loads(data.get(f'{name}_data'))


class SingleFileWidget(Widget):
    needs_multipart_form = True
    template_name = 'django_spire/file/widget/single_file_widget.html'

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def render(self, name: str, value, attrs: dict | None = None, renderer=None) -> SafeString:
        context = self.get_context(name, value, attrs)
        template = loader.get_template(self.template_name).render(context)
        return mark_safe(template)

    def value_from_datadict(self, data: dict, files, name: str) -> dict:
        return json.loads(data.get(f'{name}_data'))



from __future__ import annotations

from django.contrib import admin
from django.urls.base import reverse
from django.utils.html import format_html

from django_spire.history.viewed.models import Viewed


@admin.register(Viewed)
class ViewAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'content_object_link', 'user_link', 'created_datetime'
    )
    list_filter = ('created_datetime',)
    search_fields = ('id', 'user__first_name', 'user__last_name', 'content_type__model')
    ordering = ('-created_datetime',)

    def content_object_link(self, view: Viewed) -> str:
        url = reverse(
            f'admin:{view.content_type.app_label}_{view.content_type.model}_change',
            args=[view.object_id]
        )

        return format_html(f'<a href="{url}">{view.content_object}</a>')

    content_object_link.short_description = 'Content Object'

    def user_link(self, view: Viewed) -> str:
        url = reverse('admin:auth_user_change', args=[view.user.id])
        return format_html(f'<a href="{url}">{view.user.get_full_name()}</a>')

    user_link.short_description = 'User'



from __future__ import annotations

import pytest

from django.test import TestCase

from django_spire.utils import (
    app_is_installed,
    get_class_from_string,
    get_class_name_from_class,
)


class TestAppIsInstalled(TestCase):
    def test_installed_app_returns_true(self) -> None:
        result = app_is_installed('auth')

        assert result is True

    def test_not_installed_app_returns_false(self) -> None:
        result = app_is_installed('nonexistent_app')

        assert result is False


class TestGetClassFromString(TestCase):
    def test_imports_class(self) -> None:
        result = get_class_from_string('django.test.TestCase')

        assert result is TestCase

    def test_invalid_class_string_raises_exception(self) -> None:
        with pytest.raises(Exception, match='not a valid class string'):
            get_class_from_string('InvalidString')

    def test_nonexistent_module_raises_exception(self) -> None:
        with pytest.raises(ModuleNotFoundError):
            get_class_from_string('nonexistent.module.ClassName')


class TestGetClassNameFromClass(TestCase):
    def test_returns_full_class_path(self) -> None:
        result = get_class_name_from_class(TestCase)

        assert result == 'django.test.testcases.TestCase'

    def test_returns_string(self) -> None:
        result = get_class_name_from_class(TestCase)

        assert isinstance(result, str)

    def test_contains_module_and_class_name(self) -> None:
        result = get_class_name_from_class(TestCase)

        assert 'TestCase' in result
        assert 'django' in result



from __future__ import annotations

from django.conf import settings

from django_spire.consts import NOTIFICATION_THROTTLE_RATE_PER_MINUTE_SETTINGS_NAME


def get_throttle_sleep_time() -> float:
    return float(getattr(settings, NOTIFICATION_THROTTLE_RATE_PER_MINUTE_SETTINGS_NAME) / 60)



from __future__ import annotations

import re

from typing import TYPE_CHECKING

from django_spire.notification.choices import NotificationStatusChoices

if TYPE_CHECKING:
    from django_spire.notification.sms.models import SmsTemporaryMedia


def update_unsent_notification_status_for_deleted_temporary_media(
    temporary_media_to_delete: list[SmsTemporaryMedia]
):
    for temporary_media in temporary_media_to_delete:
        if temporary_media.has_unsent_notifications():
            temporary_media.sms_notifications.all().update(
                notification__status=NotificationStatusChoices.ERRORED,
                notification__status_message='SMS temporary media expired before notification was sent',
            )

def format_to_international_phone_number(phone_number: str, country_code: str='1') -> str:
    """
    Args: phone_number:
    Returns: international phone number format
    """

    if not phone_number:
        message = f'No phone number provided: {phone_number}'
        raise ValueError(message)

    # Remove extension numbers
    main_number = re.split(r'(?:ext\.?|x)\s*\d+', phone_number, flags=re.IGNORECASE)[0]

    # Get all digit characters
    digit_number = re.sub(r'\D', '', main_number)

    if digit_number.startswith(country_code) and len(digit_number) == 10 + len(country_code):
        digit_number = digit_number[len(country_code):]

    # Check if the number is in local format or already in international format
    if len(digit_number) == 10:
        return f'+{country_code}{digit_number}'

    if len(digit_number) == 10 + len(country_code) and digit_number.startswith(country_code):
        return f'+{digit_number}'

    message = f'Invalid phone number: {phone_number}'
    raise ValueError(message)



from __future__ import annotations

import pytest

from unittest.mock import MagicMock

from django.test import TestCase

from django_spire.core.redirect.generic_redirect import reverse_generic_relation


class TestReverseGenericRelation(TestCase):
    def test_model_not_in_map_raises_key_error(self) -> None:
        """
        Since CONTENT_OBJECT_URL_MAP is empty by default,
        any model lookup raises KeyError.
        """

        content_object = MagicMock()
        content_object.__class__.__name__ = 'UnknownModel'

        with pytest.raises(KeyError):
            reverse_generic_relation(content_object)

    def test_uses_lowercase_model_name(self) -> None:
        """Verifies the model name is lowercased for map lookup."""

        content_object = MagicMock()
        content_object.__class__.__name__ = 'MyModel'

        with pytest.raises(KeyError) as exc_info:
            reverse_generic_relation(content_object)

        assert exc_info.value.args[0] == 'mymodel'



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING

from django.core.files.uploadedfile import InMemoryUploadedFile, SimpleUploadedFile

from django_spire.file.models import File

if TYPE_CHECKING:
    from django.contrib.contenttypes.models import ContentType


def create_test_file(
    content_type: ContentType | None = None,
    object_id: int | None = None,
    name: str = 'test_file',
    file_type: str = 'pdf',
    size: str = '1.5 Mb',
    related_field: str | None = None,
    is_active: bool = True,
    is_deleted: bool = False,
) -> File:
    file_content = b'test content'
    uploaded_file = SimpleUploadedFile(
        name=f'{name}.{file_type}',
        content=file_content,
        content_type='application/octet-stream'
    )

    return File.objects.create(
        content_type=content_type,
        object_id=object_id,
        file=uploaded_file,
        name=name,
        type=file_type,
        size=size,
        related_field=related_field,
        is_active=is_active,
        is_deleted=is_deleted,
    )


def create_test_in_memory_uploaded_file(
    name: str = 'test_file',
    file_type: str = 'pdf',
    content: bytes = b'test content',
) -> InMemoryUploadedFile:
    file_io = BytesIO(content)
    file_io.seek(0)

    return InMemoryUploadedFile(
        file=file_io,
        field_name='file',
        name=f'{name}.{file_type}',
        content_type='application/octet-stream',
        size=len(content),
        charset=None,
    )



from __future__ import annotations

from django.db.models import QuerySet


class CommentQuerySet(QuerySet):
    def active(self):
        return self.filter(is_deleted=False)

    def top_level(self):
        return self.filter(parent__isnull=True)

    def prefetch_user(self):
        return self.prefetch_related('user')

    def prefetch_parent(self):
        return self.prefetch_related('parent')

    def prefetch_replies(self):
        return self.prefetch_related('children')



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.enums import TicketEventType
from django_spire.help_desk.exceptions import (
    HelpDeskNotificationRecipientMissingEmailError,
    TicketEventNotificationTypeNotSupportedError,
)
from django_spire.notification.choices import NotificationTypeChoices


class HelpDeskExceptionsTests(BaseTestCase):
    def test_ticket_event_notification_type_not_supported_error_is_type_error(self):
        assert issubclass(TicketEventNotificationTypeNotSupportedError, TypeError)

    def test_ticket_event_notification_type_not_supported_error_message(self):
        error = TicketEventNotificationTypeNotSupportedError(
            TicketEventType.NEW,
            NotificationTypeChoices.SMS
        )
        assert 'Event type' in str(error)
        assert 'Notification type' in str(error)

    def test_ticket_event_notification_type_not_supported_error_message_without_event_type(self):
        error = TicketEventNotificationTypeNotSupportedError(
            None,
            NotificationTypeChoices.SMS
        )
        assert 'Notification type not supported' in str(error)

    def test_helpdesk_notification_recipient_missing_email_error_is_value_error(self):
        assert issubclass(HelpDeskNotificationRecipientMissingEmailError, ValueError)

    def test_helpdesk_notification_recipient_missing_email_error_message(self):
        error = HelpDeskNotificationRecipientMissingEmailError('John Doe')
        assert 'John Doe' in str(error)
        assert 'missing an email address' in str(error)



from __future__ import annotations

from dandy import Prompt
from django.forms import model_to_dict

from django_spire.ai.context.models import Organization


def organization_info_prompt() -> Prompt:
    org = Organization.objects.get_only_or_none()

    if org is None:
        return Prompt().text('There is no organization information available.')

    org_dict = model_to_dict(
        org, exclude=['id', 'created_datetime', 'is_active', 'is_deleted']
    )

    return Prompt().heading('Organization Information').dict(org_dict)



from __future__ import annotations

from django.test import TestCase

from django_spire.contrib.seeding.field.static import StaticFieldSeeder


class TestStaticFieldSeeder(TestCase):
    def test_seeds_static_fields(self) -> None:
        fields = {
            'category': ('static', 'books'),
            'in_stock': ('static', True)
        }
        seeder = StaticFieldSeeder(fields=fields)

        result = seeder.seed(count=3)

        assert len(result) == 3

        for row in result:
            assert row['category'] == 'books'
            assert row['in_stock'] is True



from __future__ import annotations

from django_spire.exceptions import DjangoSpireError


class AppNotificationError(DjangoSpireError):
    pass


class MissingUserError(AppNotificationError):
    def __init__(self) -> None:
        super().__init__('AppNotifications must have a user associated with them')



from __future__ import annotations

from django.contrib.auth.models import User
from django.db import models
from django.utils.formats import localize
from django.utils.timezone import now

from django_spire.ai.mixins import AiUsageMixin


class AiUsage(AiUsageMixin):
    recorded_date = models.DateField(default=now)

    def __str__(self):
        return f'"{self.recorded_date}" ai usage'

    class Meta:
        db_table = 'django_spire_ai_usage'
        verbose_name = 'AI Usage'
        verbose_name_plural = 'AI Usage'


class AiInteraction(AiUsageMixin):
    ai_usage = models.ForeignKey(
        AiUsage,
        on_delete=models.CASCADE,
        related_name='interactions',
        related_query_name='interaction',
    )

    user = models.ForeignKey(
        User,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='ai_interactions',
        related_query_name='ai_interaction',
    )

    actor = models.CharField(max_length=100)
    user_email = models.EmailField(blank=True, null=True)
    user_first_name = models.CharField(max_length=100, blank=True, null=True)
    user_last_name = models.CharField(max_length=100, blank=True, null=True)

    module_name = models.TextField()
    callable_name = models.TextField()

    exception = models.TextField(blank=True, null=True)
    stack_trace = models.TextField(blank=True, null=True)

    created_datetime = models.DateTimeField(default=now)

    interaction = models.JSONField(blank=True, null=True)

    def __str__(self):
        return f'"{self.actor}" interaction on "{localize(self.created_datetime)}"'

    def save(self, *args, **kwargs) -> None:
        if self.user:
            if self.actor is None:
                self.actor = self.user.get_full_name()

            if self.user_email is None:
                self.user_email = self.user.email

            if self.user_first_name is None:
                self.user_first_name = self.user.first_name

            if self.user_last_name is None:
                self.user_last_name = self.user.last_name

        super().save(*args, **kwargs)

    class Meta:
        db_table = 'django_spire_ai_interaction'
        verbose_name = 'AI Interaction'
        verbose_name_plural = 'AI Interactions'



from __future__ import annotations

from typing import TYPE_CHECKING

from django.shortcuts import get_object_or_404

from django_spire.auth.permissions.decorators import permission_required
from django_spire.auth.permissions.tools import generate_group_perm_data, generate_user_perm_data
from django_spire.contrib.pagination.pagination import paginate_list
from django_spire.auth.user.models import AuthUser
from django_spire.contrib.generic_views import portal_views

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.template.response import TemplateResponse


@permission_required('django_spire_auth_user.view_authuser')
def detail_view(request: WSGIRequest, pk: int) -> TemplateResponse:
    user = get_object_or_404(AuthUser, pk=pk)
    group_list = user.groups.all()

    context_data = {
        'user': user,
        'group_list': group_list,
        'group_list_permission_data': [generate_group_perm_data(group) for group in group_list],
        'user_perm_data': generate_user_perm_data(user)
    }

    return portal_views.detail_view(
        request,
        context_data=context_data,
        obj=user,
        template='django_spire/auth/user/page/detail_page.html'
    )


@permission_required('django_spire_auth_user.view_authuser')
def list_view(request: WSGIRequest) -> TemplateResponse:
    active_user_list = AuthUser.objects.filter(is_active=True).prefetch_related('groups').order_by('first_name', 'last_name')
    inactive_user_list = AuthUser.objects.filter(is_active=False).prefetch_related('groups').order_by('first_name', 'last_name')

    context_data = {
        'active_user_list': paginate_list(active_user_list, page_number=request.GET.get('page', 1)),
        'inactive_user_list': paginate_list(inactive_user_list, page_number=request.GET.get('page', 1))
    }

    return portal_views.list_view(
        request,
        context_data=context_data,
        model=AuthUser,
        template='django_spire/auth/user/page/list_page.html'
    )



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.exceptions import DjangoSpireError
from django_spire.notification.app.exceptions import AppNotificationError
from django_spire.notification.email.exceptions import EmailNotificationError
from django_spire.notification.exceptions import NotificationError
from django_spire.notification.sms.exceptions import (
    SmsNotificationError,
    SmsTemporaryMediaError,
    TwilioAPIConcurrentError,
    TwilioError,
)


class NotificationExceptionsTests(BaseTestCase):
    def test_notification_error_is_django_spire_error(self):
        assert issubclass(NotificationError, DjangoSpireError)

    def test_notification_error_message(self):
        error = NotificationError('Test error')
        assert str(error) == 'Test error'


class AppNotificationExceptionsTests(BaseTestCase):
    def test_app_notification_error_is_django_spire_error(self):
        assert issubclass(AppNotificationError, DjangoSpireError)

    def test_app_notification_error_message(self):
        error = AppNotificationError('Test error')
        assert str(error) == 'Test error'


class EmailNotificationExceptionsTests(BaseTestCase):
    def test_email_notification_error_is_django_spire_error(self):
        assert issubclass(EmailNotificationError, DjangoSpireError)

    def test_email_notification_error_message(self):
        error = EmailNotificationError('Test error')
        assert str(error) == 'Test error'


class SmsNotificationExceptionsTests(BaseTestCase):
    def test_sms_notification_error_is_django_spire_error(self):
        assert issubclass(SmsNotificationError, DjangoSpireError)

    def test_sms_notification_error_message(self):
        error = SmsNotificationError('Test error')
        assert str(error) == 'Test error'

    def test_sms_temporary_media_error_is_sms_notification_error(self):
        assert issubclass(SmsTemporaryMediaError, SmsNotificationError)

    def test_twilio_error_is_exception(self):
        assert issubclass(TwilioError, Exception)

    def test_twilio_api_concurrent_error_is_twilio_error(self):
        assert issubclass(TwilioAPIConcurrentError, TwilioError)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.files.base import ContentFile

from django_spire.file.models import File
from django_spire.file.utils import random_64_char_token

if TYPE_CHECKING:
    from django.core.files.uploadedfile import InMemoryUploadedFile

    from django_spire.file.mixins import FileModelMixin


@dataclass
class FileFormatter:
    file: InMemoryUploadedFile
    related_field: str = None
    name: str = ...
    type: str = ...
    app_name: str = 'Uncategorized'

    def __post_init__(self) -> None:
        self.name = self._format_name()
        self.type = self._parse_type()

    def _format_name(self) -> str:
        return self.file.name.rsplit('.', 1)[0]

    def _parse_type(self) -> str:
        return self.file.name.rsplit('.', 1)[1]

    @property
    def location(self) -> str:
        location = settings.BASE_FOLDER_NAME + '/'
        location += str(self.app_name) + '/'

        if self.related_field is not None:
            location += str(self.related_field) + '/'

        location += random_64_char_token() + '/'
        location += self.name

        return location

    def null_file_obj(self) -> File:
        file_obj = File(
            name=self.name,
            size=self.size_verbose(),
            type=self.type,
            related_field=self.related_field,
        )
        self.save_file(file_obj)

        return file_obj

    def save_file(self, file_obj: File) -> None:
        file_path = self.location + '.' + self.type
        file_obj.file.save(file_path, ContentFile(self.file.read()), save=False)

    def size_verbose(self) -> str:
        if self.file.size < 512000:
            value = round(self.file.size / 1000, 2)
            ext = ' kb'
        elif self.file.size < 512000 * 1000:
            value = round(self.file.size / 1000000, 2)
            ext = ' Mb'
        elif self.file.size < 512000 * 1000 * 1000:
            value = round(self.file.size / 1000000000, 2)
            ext = ' Gb'
        else:
            value = ''
            ext = '>1.00 Tb'

        return str(value) + ext


@dataclass
class FileContentObjectFormatter(FileFormatter):
    content_object: FileModelMixin = ...

    @property
    def location(self) -> str:
        location = settings.BASE_FOLDER_NAME + '/'
        location += str(self.content_object._meta.app_label) + '/'

        if self.related_field is not None:
            location += str(self.related_field) + '/'

        location += random_64_char_token() + '/'
        location += self.name
        return location

    def null_file_obj(self) -> File:
        file_obj = File(
            content_type=ContentType.objects.get_for_model(self.content_object),
            object_id=self.content_object.id,
            name=self.name,
            size=self.size_verbose(),
            type=self.type,
            related_field=self.related_field,
        )
        self.save_file(file_obj)

        return file_obj


@dataclass
class FileUploader(ABC):
    related_field: str | None
    app_name: str = 'Uncategorized'

    def null_file_obj(self, file: InMemoryUploadedFile) -> File:
        formatted_file = FileFormatter(
            file=file,
            related_field=self.related_field,
            app_name=self.app_name
        )
        return formatted_file.null_file_obj()

    @abstractmethod
    def upload(self, *args, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    def upload_from_form_field(self, *args, **kwargs) -> None:
        raise NotImplementedError


@dataclass
class SingleFileUploader(FileUploader):
    def delete_old_files(self, content_object: FileModelMixin) -> None:
        old_files = content_object.files.active().related_field(self.related_field)

        for file in old_files:
            file.is_deleted = True
            file.is_active = False

        File.objects.bulk_update(old_files, ['is_active', 'is_deleted'])

    def upload(self, file: InMemoryUploadedFile) -> File:
        file = self.null_file_obj(file)
        file.save()
        return file

    def upload_from_form_field(
        self,
        form_data: dict,
        content_object: FileModelMixin
    ) -> File:
        try:
            return content_object.files.get(id=form_data['id'])
        except ObjectDoesNotExist:
            self.delete_old_files(content_object)
            file = File.objects.get(id=form_data['id'])
            file.content_type = ContentType.objects.get_for_model(content_object.__class__)
            file.object_id = content_object.id
            file.related_field = self.related_field
            return file.save()


@dataclass
class MultiFileUploader(FileUploader):
    def upload(self, files: list[InMemoryUploadedFile]) -> list[File]:
        files_to_upload = [self.null_file_obj(file) for file in files]
        return File.objects.bulk_create(files_to_upload)

    def upload_from_form_field(
        self,
        form_data: list[dict],
        content_object: FileModelMixin
    ) -> list[File]:
        file_ids = [file['id'] for file in form_data]

        # Delete files that are no longer in the form data
        old_files = (
            content_object
            .files
            .active()
            .related_field(self.related_field)
            .exclude(id__in=file_ids)
        )

        for old_file in old_files:
            old_file.is_deleted = True
            old_file.is_active = False

        File.objects.bulk_update(old_files, ['is_active', 'is_deleted'])

        link_files = File.objects.filter(id__in=file_ids, object_id__isnull=True)
        content_type = ContentType.objects.get_for_model(content_object.__class__)

        for link_file in link_files:
            link_file.content_type = content_type
            link_file.object_id = content_object.id
            link_file.related_field = self.related_field

        File.objects.bulk_update(
            link_files,
            fields=['content_type', 'object_id', 'related_field']
        )

        return content_object.files.active().related_field(self.related_field)



from __future__ import annotations

from django_spire.ai.chat.message_intel import BaseMessageIntel
from django_spire.knowledge.intelligence.intel.answer_intel import AnswerIntel
from django_spire.knowledge.intelligence.intel.entry_intel import EntriesIntel


class KnowledgeMessageIntel(BaseMessageIntel):
    _template: str = 'django_spire/knowledge/message/knowledge_message_intel.html'
    answer_intel: AnswerIntel
    entries_intel: EntriesIntel

    def render_to_str(self) -> str:
        return self.answer_intel.answer



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.entry.version.block.choices import BlockTypeChoices
from django_spire.knowledge.entry.version.block.models import EntryVersionBlock
from django_spire.knowledge.entry.version.tests.factories import create_test_entry_version


class EntryVersionBlockFactoryServiceTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.entry_version = create_test_entry_version()

    def test_create_validated_block_text(self):
        block = EntryVersionBlock.services.factory.create_validated_block(
            entry_version=self.entry_version,
            type=BlockTypeChoices.TEXT,
            data={'text': 'Test content'},
            order=0
        )
        assert block.type == BlockTypeChoices.TEXT
        assert block.order == 0
        assert block._block_data == {'text': 'Test content'}

    def test_create_validated_block_heading(self):
        block = EntryVersionBlock.services.factory.create_validated_block(
            entry_version=self.entry_version,
            type=BlockTypeChoices.HEADING,
            data={'text': 'Title', 'level': 1},
            order=0
        )
        assert block.type == BlockTypeChoices.HEADING
        assert block._block_data['level'] == 1

    def test_create_validated_block_list(self):
        block = EntryVersionBlock.services.factory.create_validated_block(
            entry_version=self.entry_version,
            type=BlockTypeChoices.LIST,
            data={
                'style': 'unordered',
                'items': [{'content': 'Item', 'items': []}]
            },
            order=0
        )
        assert block.type == BlockTypeChoices.LIST

    def test_create_validated_block_with_tunes(self):
        tunes = {'alignment': 'center'}
        block = EntryVersionBlock.services.factory.create_validated_block(
            entry_version=self.entry_version,
            type=BlockTypeChoices.TEXT,
            data={'text': 'Centered'},
            order=0,
            tunes=tunes
        )
        assert block.tunes == tunes

    def test_create_validated_block_sets_version(self):
        block = EntryVersionBlock.services.factory.create_validated_block(
            entry_version=self.entry_version,
            type=BlockTypeChoices.TEXT,
            data={'text': 'Test'},
            order=0
        )
        assert block.version == self.entry_version



from __future__ import annotations

from django_spire.ai.prompt.tuning.prompts import (
    duplication_removal_bot_instruction_prompt,
    example_optimization_bot_instruction_prompt,
    formatting_bot_instruction_prompt,
    instruction_clarity_bot_instruction_prompt,
    persona_bot_instruction_prompt,
    prompt_tuning_input_prompt,
    prompt_tuning_instruction_bot_prompt,
    specialized_bot_input_prompt,
)
from django_spire.ai.prompt.tuning.choices import OutcomeRatingChoices
from django_spire.ai.prompt.tuning.intel import PromptTestingIntel, PromptTuningIntel
from django_spire.core.tests.test_cases import BaseTestCase


class OutcomeRatingChoicesTests(BaseTestCase):
    def test_terrible_rating(self) -> None:
        assert OutcomeRatingChoices.TERRIBLE.value == 1
        assert OutcomeRatingChoices.TERRIBLE.label == 'terrible'

    def test_bad_rating(self) -> None:
        assert OutcomeRatingChoices.BAD.value == 2
        assert OutcomeRatingChoices.BAD.label == 'bad'

    def test_ok_rating(self) -> None:
        assert OutcomeRatingChoices.OK.value == 3
        assert OutcomeRatingChoices.OK.label == 'ok'

    def test_good_rating(self) -> None:
        assert OutcomeRatingChoices.GOOD.value == 4
        assert OutcomeRatingChoices.GOOD.label == 'good'

    def test_great_rating(self) -> None:
        assert OutcomeRatingChoices.GREAT.value == 5
        assert OutcomeRatingChoices.GREAT.label == 'great'

    def test_amazing_rating(self) -> None:
        assert OutcomeRatingChoices.AMAZING.value == 6
        assert OutcomeRatingChoices.AMAZING.label == 'amazing'

    def test_all_ratings_count(self) -> None:
        assert len(OutcomeRatingChoices) == 6


class PromptTuningIntelTests(BaseTestCase):
    def test_prompt_tuning_intel_creation(self) -> None:
        intel = PromptTuningIntel(prompt='Test prompt')

        assert intel.prompt == 'Test prompt'

    def test_prompt_tuning_intel_empty_prompt(self) -> None:
        intel = PromptTuningIntel(prompt='')

        assert intel.prompt == ''

    def test_prompt_tuning_intel_long_prompt(self) -> None:
        long_prompt = 'A' * 10000
        intel = PromptTuningIntel(prompt=long_prompt)

        assert intel.prompt == long_prompt
        assert len(intel.prompt) == 10000


class PromptTestingIntelTests(BaseTestCase):
    def test_prompt_testing_intel_creation(self) -> None:
        intel = PromptTestingIntel(result='Test result')

        assert intel.result == 'Test result'

    def test_prompt_testing_intel_empty_result(self) -> None:
        intel = PromptTestingIntel(result='')

        assert intel.result == ''


class PromptTuningPromptsTests(BaseTestCase):
    def test_prompt_tuning_instruction_bot_prompt(self) -> None:
        prompt = prompt_tuning_instruction_bot_prompt()

        assert prompt is not None
        assert 'System Prompt Tuning Expert' in prompt.to_str()

    def test_prompt_tuning_input_prompt(self) -> None:
        prompt = prompt_tuning_input_prompt('Test system prompt', 'Test feedback')

        assert prompt is not None
        assert 'Test system prompt' in prompt.to_str()
        assert 'Test feedback' in prompt.to_str()

    def test_formatting_bot_instruction_prompt(self) -> None:
        prompt = formatting_bot_instruction_prompt()

        assert prompt is not None
        assert 'Prompt Formatting Expert' in prompt.to_str()

    def test_instruction_clarity_bot_instruction_prompt(self) -> None:
        prompt = instruction_clarity_bot_instruction_prompt()

        assert prompt is not None
        assert 'Instruction Clarity Expert' in prompt.to_str()

    def test_persona_bot_instruction_prompt(self) -> None:
        prompt = persona_bot_instruction_prompt()

        assert prompt is not None
        assert 'Persona Consistency Expert' in prompt.to_str()

    def test_duplication_removal_bot_instruction_prompt(self) -> None:
        prompt = duplication_removal_bot_instruction_prompt()

        assert prompt is not None
        assert 'Duplication Removal Expert' in prompt.to_str()

    def test_example_optimization_bot_instruction_prompt(self) -> None:
        prompt = example_optimization_bot_instruction_prompt()

        assert prompt is not None
        assert 'Example Optimization Expert' in prompt.to_str()

    def test_specialized_bot_input_prompt(self) -> None:
        prompt = specialized_bot_input_prompt('Test system prompt')

        assert prompt is not None
        assert 'Test system prompt' in prompt.to_str()



from __future__ import annotations

from django_spire.contrib.progress.enums import ProgressStatus
from django_spire.contrib.progress.tasks import ParallelTask, SequentialTask
from django_spire.contrib.progress.session import ProgressSession


__all__ = [
    'ParallelTask',
    'ProgressSession',
    'ProgressStatus',
    'SequentialTask',
]



from __future__ import annotations

from django_spire.ai.chat.intelligence.prompts import chat_system_prompt
from django_spire.ai.context.models import Organization
from django_spire.core.tests.test_cases import BaseTestCase


class ChatPromptsTests(BaseTestCase):
    def test_chat_system_prompt_returns_prompt(self) -> None:
        prompt = chat_system_prompt()

        assert prompt is not None

    def test_chat_system_prompt_contains_assistant_text(self) -> None:
        prompt = chat_system_prompt()
        prompt_str = prompt.to_str()

        assert 'chat assistant' in prompt_str.lower()

    def test_chat_system_prompt_contains_rules(self) -> None:
        prompt = chat_system_prompt()
        prompt_str = prompt.to_str()

        assert 'rules' in prompt_str.lower()

    def test_chat_system_prompt_with_organization(self) -> None:
        Organization.objects.create(
            name='Test Company',
            description='A test company for testing purposes'
        )

        prompt = chat_system_prompt()
        prompt_str = prompt.to_str()

        assert 'organization' in prompt_str.lower()

    def test_chat_system_prompt_without_organization(self) -> None:
        prompt = chat_system_prompt()
        prompt_str = prompt.to_str()

        assert prompt_str is not None
        assert len(prompt_str) > 0

    def test_chat_system_prompt_ai_illusion_rule(self) -> None:
        prompt = chat_system_prompt()
        prompt_str = prompt.to_str()

        assert 'ai' in prompt_str.lower() or 'illusion' in prompt_str.lower()



from __future__ import annotations

from abc import ABC, abstractmethod

from dandy import BaseIntel
from django.template.loader import render_to_string


class BaseMessageIntel(BaseIntel, ABC):
    _template: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        template = cls.__dict__.get('_template')

        if template is None and hasattr(cls, '__private_attributes__'):
            private_attr = cls.__private_attributes__.get('_template')

            if private_attr and hasattr(private_attr, 'default'):
                template = private_attr.default

        if not template:
            message = f'{cls.__module__}.{cls.__qualname__}._template must be set'
            raise ValueError(message)

    @abstractmethod
    def render_to_str(self) -> str:
        raise NotImplementedError

    def render_template_to_str(self, context_data: dict | None = None):
        return render_to_string(
            template_name=self._template,
            context={**self.model_dump(), **(context_data or {})},
        )

    @property
    def template(self) -> str:
        return self._template


class DefaultMessageIntel(BaseMessageIntel):
    _template: str = 'django_spire/ai/chat/message/default_message.html'
    text: str

    def render_to_str(self) -> str:
        return self.text



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.handlers.wsgi import WSGIRequest

from django_spire.ai.chat.message_intel import DefaultMessageIntel, BaseMessageIntel
from django_spire.core.tag.intelligence.tag_set_bot import TagSetBot
from django_spire.knowledge.collection.models import Collection
from django_spire.knowledge.intelligence.bots.knowledge_answer_bot import KnowledgeAnswerBot
from django_spire.knowledge.intelligence.bots.knowledge_entries_bot import KnowledgeEntriesBot
from django_spire.knowledge.intelligence.intel.message_intel import KnowledgeMessageIntel

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from dandy.llm.request.message import MessageHistory


NO_KNOWLEDGE_MESSAGE_INTEL = DefaultMessageIntel(
    text='Sorry, I could not find any information on that.'
)


def knowledge_search_workflow(
    request: WSGIRequest,
    user_input: str,
    message_history: MessageHistory | None = None
) -> BaseMessageIntel | None:
    user_tag_set = TagSetBot().process(user_input)

    def get_top_scored_from_dict_to_list(
            scored_dict: dict[str, float],
            score_floor: float = 0.05
    ) -> list:
        if not scored_dict:
            return []

        min_score = min(scored_dict.values())
        max_score = max(scored_dict.values())

        if min_score == 0 and max_score == 0:
            return []

        median_score = (max_score - min_score) / 2

        top_scored_list = []

        for key, value in scored_dict.items():
            if value >= score_floor and value >= median_score:
                top_scored_list.append(key)

        return top_scored_list

    collections = get_top_scored_from_dict_to_list({
        collection: collection.services.tag.get_score_percentage_from_aggregated_tag_set_weighted(
            user_tag_set
        )
        for collection in Collection.objects.all().annotate_entry_count()
    })

    entries = get_top_scored_from_dict_to_list({
        entry: entry.services.tag.get_score_percentage_from_tag_set_weighted(user_tag_set)
        for collection in collections
        for entry in collection.entries.all()
    })

    if not entries:
        return NO_KNOWLEDGE_MESSAGE_INTEL

    answer_intel_future = KnowledgeAnswerBot(llm_temperature=0.5).process_to_future(
        user_input=user_input,
        entries=entries
    )

    entries_intel_future = KnowledgeEntriesBot(llm_temperature=0.5).process_to_future(
        user_input=user_input,
        entries=entries
    )

    return KnowledgeMessageIntel(
        answer_intel=answer_intel_future.result,
        entries_intel=entries_intel_future.result,
    )



from __future__ import annotations

from unittest.mock import MagicMock, patch

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.automations import process_notifications


class NotificationAutomationsTests(BaseTestCase):
    @patch('django_spire.notification.automations.NotificationManager')
    def test_process_notifications_calls_manager(self, mock_manager_class: MagicMock):
        mock_manager = MagicMock()
        mock_manager_class.return_value = mock_manager

        result = process_notifications()

        mock_manager.process_ready_notifications.assert_called_once()
        assert result == 'Successfully Completed'



from __future__ import annotations

import json

from typing import Sequence

from django import template


register = template.Library()


@register.filter(name='to_json')
def to_json(value: dict | Sequence) -> str:
    try:
        return json.dumps(value)
    except (TypeError, ValueError):
        return ''



from __future__ import annotations

from django.apps import AppConfig

from django_spire.utils import check_required_apps


class KnowledgeConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    label = 'django_spire_knowledge'
    name = 'django_spire.knowledge'
    MODEL_PERMISSIONS = (
        {
            'name': 'knowledge',
            'verbose_name': 'Knowledge',
            'model_class_path': 'django_spire.knowledge.collection.models.Collection',
            'is_proxy_model': False,
        },
    )

    REQUIRED_APPS = ('django_spire_core',)

    URLPATTERNS_INCLUDE = 'django_spire.knowledge.urls'
    URLPATTERNS_NAMESPACE = 'knowledge'

    def ready(self) -> None:
        check_required_apps(self.label)



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.exceptions import DjangoSpireError

if TYPE_CHECKING:
    from django_spire.notification.choices import NotificationTypeChoices


class NotificationError(DjangoSpireError):
    pass


class InvalidNotificationTypeError(NotificationError):
    def __init__(self, expected_type: NotificationTypeChoices, actual_type: NotificationTypeChoices) -> None:
        super().__init__(
            f'Expected notification type {expected_type}, '
            f'but received {actual_type}'
        )



from __future__ import annotations

from django.contrib import admin
from django.urls import reverse
from django.utils.html import format_html

from django_spire.file import models


@admin.register(models.File)
class FileAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'name', 'type', 'size', 'content_object_link', 'file_link'
    )
    list_filter = ('type',)
    search_fields = ('id', 'name', 'type')
    ordering = ('-id',)

    def content_object_link(self, file: models.File) -> str:
        if file.content_object:
            url = reverse(
                f'admin:{file.content_type.app_label}_{file.content_type.model}_change',
                args=[file.object_id]
            )

            return format_html('<a href="{}">{}</a>', url, file.content_object)

        return 'No Related Object'

    content_object_link.short_description = 'Content Object'

    def file_link(self, file: models.File) -> str:
        return format_html('<a href="{}" download>{}</a>', file.file.url, file.name)

    file_link.short_description = 'File Download Link'



# Generated by Django 5.2.7 on 2025-11-09 05:02

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(max_length=128, unique=True)),
            ],
            options={
                'verbose_name': 'Tag',
                'verbose_name_plural': 'Tags',
                'db_table': 'django_spire_core_tag',
            },
        ),
    ]



from __future__ import annotations

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.permissions.permissions import (
    GroupPermissions,
    ModelPermission,
    ModelPermissions,
    UserPermissionHelper,
)
from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase


class ModelPermissionTestCase(BaseTestCase):
    def test_model_permission_attributes(self) -> None:
        mp = ModelPermission(
            name='test',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True,
            verbose_name='Test Permission'
        )
        assert mp.name == 'test'
        assert mp.model_class_path == 'django_spire.auth.group.models.AuthGroup'
        assert mp.is_proxy_model
        assert mp.verbose_name == 'Test Permission'

    def test_model_class_property(self) -> None:
        mp = ModelPermission(
            name='group',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )
        assert mp.model_class == AuthGroup

    def test_model_permission_without_verbose_name(self) -> None:
        mp = ModelPermission(
            name='test',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )
        assert mp.verbose_name is None

    def test_model_permission_non_proxy(self) -> None:
        mp = ModelPermission(
            name='test',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=False
        )
        assert not mp.is_proxy_model

    def test_model_permission_name_case_preserved(self) -> None:
        mp = ModelPermission(
            name='TestName',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )
        assert mp.name == 'TestName'

    def test_model_permission_verbose_name_unicode(self) -> None:
        mp = ModelPermission(
            name='test',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True,
            verbose_name='Tëst Përmission'
        )
        assert mp.verbose_name == 'Tëst Përmission'


class ModelPermissionsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.model_permission = ModelPermission(
            name='group',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )
        self.model_permissions = ModelPermissions(self.model_permission)

    def test_app_label(self) -> None:
        assert self.model_permissions.app_label == 'django_spire_auth_group'

    def test_model_name(self) -> None:
        assert self.model_permissions.model_name == 'authgroup'

    def test_permissions_queryset(self) -> None:
        assert self.model_permissions.permissions.count() > 0

    def test_find_permissions_by_level_none(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(0)
        assert len(perms) == 0

    def test_find_permissions_by_level_view(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(1)
        codenames = [p.codename for p in perms]
        assert 'view_authgroup' in codenames

    def test_find_permissions_by_level_add(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(2)
        codenames = [p.codename for p in perms]
        assert 'view_authgroup' in codenames
        assert 'add_authgroup' in codenames

    def test_find_permissions_by_level_change(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(3)
        codenames = [p.codename for p in perms]
        assert 'view_authgroup' in codenames
        assert 'add_authgroup' in codenames
        assert 'change_authgroup' in codenames

    def test_find_permissions_by_level_delete(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(4)
        codenames = [p.codename for p in perms]
        assert 'view_authgroup' in codenames
        assert 'add_authgroup' in codenames
        assert 'change_authgroup' in codenames
        assert 'delete_authgroup' in codenames

    def test_special_role_list(self) -> None:
        roles = self.model_permissions.special_role_list()
        for role in roles:
            assert role.codename.startswith('can')

    def test_get_special_role_not_found(self) -> None:
        result = self.model_permissions.get_special_role('nonexistent_role')
        assert result is None

    def test_permissions_returns_queryset(self) -> None:
        assert hasattr(self.model_permissions.permissions, 'filter')

    def test_model_attribute(self) -> None:
        assert self.model_permissions.model == AuthGroup

    def test_is_proxy_model_attribute(self) -> None:
        assert self.model_permissions.is_proxy_model

    def test_find_permissions_by_level_returns_list(self) -> None:
        perms = self.model_permissions.find_permissions_by_level(2)
        assert isinstance(perms, list)

    def test_special_role_list_returns_list(self) -> None:
        roles = self.model_permissions.special_role_list()
        assert isinstance(roles, list)


class GroupPermissionsTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')
        self.model_permission = ModelPermission(
            name='group',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )
        self.group_permissions = GroupPermissions(self.group, self.model_permission)

    def test_initial_perm_level_is_zero(self) -> None:
        assert self.group_permissions.perm_level() == 0

    def test_perm_level_verbose_none(self) -> None:
        assert self.group_permissions.perm_level_verbose() == 'None'

    def test_update_perms_to_view(self) -> None:
        self.group_permissions.update_perms(1)
        assert self.group_permissions.perm_level() == 1
        assert self.group_permissions.perm_level_verbose() == 'View'

    def test_update_perms_to_add(self) -> None:
        self.group_permissions.update_perms(2)
        assert self.group_permissions.perm_level() == 2
        assert self.group_permissions.perm_level_verbose() == 'Add'

    def test_update_perms_to_change(self) -> None:
        self.group_permissions.update_perms(3)
        assert self.group_permissions.perm_level() == 3
        assert self.group_permissions.perm_level_verbose() == 'Change'

    def test_update_perms_to_delete(self) -> None:
        self.group_permissions.update_perms(4)
        assert self.group_permissions.perm_level() == 4
        assert self.group_permissions.perm_level_verbose() == 'Delete'

    def test_update_perms_downgrades(self) -> None:
        self.group_permissions.update_perms(4)
        self.group_permissions.update_perms(1)
        assert self.group_permissions.perm_level() == 1

    def test_update_perms_to_none(self) -> None:
        self.group_permissions.update_perms(4)
        self.group_permissions.update_perms(0)
        assert self.group_permissions.perm_level() == 0

    def test_remove_special_permissions(self) -> None:
        content_type = ContentType.objects.get_for_model(AuthGroup)
        perms = Permission.objects.filter(content_type=content_type)
        filtered = GroupPermissions.remove_special_permissions(perms)
        for perm in filtered:
            assert not perm.codename.startswith('can')

    def test_set_group_perms_updates(self) -> None:
        self.group_permissions.update_perms(2)
        self.group_permissions.set_group_perms()
        assert self.group_permissions.perm_level() == 2

    def test_has_special_role_false(self) -> None:
        assert not self.group_permissions.has_special_role('nonexistent')

    def test_group_attribute(self) -> None:
        assert self.group_permissions.group == self.group

    def test_model_permissions_attribute(self) -> None:
        assert self.group_permissions.model_permissions is not None

    def test_group_perms_attribute(self) -> None:
        assert self.group_permissions.group_perms is not None

    def test_update_perms_multiple_times(self) -> None:
        for level in [1, 2, 3, 4, 3, 2, 1, 0]:
            self.group_permissions.update_perms(level)
            assert self.group_permissions.perm_level() == level

    def test_perm_level_returns_int(self) -> None:
        level = self.group_permissions.perm_level()
        assert isinstance(level, int)

    def test_perm_level_verbose_returns_str(self) -> None:
        verbose = self.group_permissions.perm_level_verbose()
        assert isinstance(verbose, str)

    def test_remove_special_permissions_with_list(self) -> None:
        content_type = ContentType.objects.get_for_model(AuthGroup)
        perms = list(Permission.objects.filter(content_type=content_type))
        filtered = GroupPermissions.remove_special_permissions(perms)
        assert isinstance(filtered, list)


class UserPermissionHelperTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.user = create_user(username='testuser')
        self.group = AuthGroup.objects.create(name='Test Group')
        self.model_permission = ModelPermission(
            name='group',
            model_class_path='django_spire.auth.group.models.AuthGroup',
            is_proxy_model=True
        )

    def test_initial_perm_level_is_zero(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 0

    def test_perm_level_verbose_none(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level_verbose() == 'None'

    def test_user_inherits_group_permissions(self) -> None:
        self.user.groups.add(self.group)
        group_perms = GroupPermissions(self.group, self.model_permission)
        group_perms.update_perms(3)

        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 3
        assert helper.perm_level_verbose() == 'Change'

    def test_user_inherits_highest_permission(self) -> None:
        group2 = AuthGroup.objects.create(name='Group 2')
        self.user.groups.add(self.group, group2)

        group_perms1 = GroupPermissions(self.group, self.model_permission)
        group_perms1.update_perms(1)

        group_perms2 = GroupPermissions(group2, self.model_permission)
        group_perms2.update_perms(4)

        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 4

    def test_user_no_groups_no_permissions(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 0

    def test_get_user_perms_returns_queryset(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        perms = helper.get_user_perms()
        assert perms is not None

    def test_user_attribute(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.user == self.user

    def test_model_permissions_attribute(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.model_permissions is not None

    def test_user_perms_attribute(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.user_perms is not None

    def test_perm_level_returns_int(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        level = helper.perm_level()
        assert isinstance(level, int)

    def test_perm_level_verbose_returns_str(self) -> None:
        helper = UserPermissionHelper(self.user, self.model_permission)
        verbose = helper.perm_level_verbose()
        assert isinstance(verbose, str)

    def test_user_removed_from_group_loses_permissions(self) -> None:
        self.user.groups.add(self.group)
        group_perms = GroupPermissions(self.group, self.model_permission)
        group_perms.update_perms(3)

        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 3

        self.user.groups.remove(self.group)
        helper2 = UserPermissionHelper(self.user, self.model_permission)
        assert helper2.perm_level() == 0

    def test_multiple_groups_same_permission_level(self) -> None:
        group2 = AuthGroup.objects.create(name='Group 2')
        self.user.groups.add(self.group, group2)

        group_perms1 = GroupPermissions(self.group, self.model_permission)
        group_perms1.update_perms(2)

        group_perms2 = GroupPermissions(group2, self.model_permission)
        group_perms2.update_perms(2)

        helper = UserPermissionHelper(self.user, self.model_permission)
        assert helper.perm_level() == 2



from __future__ import annotations

from keyring.testing.util import random_string

from django_spire.auth.user.tests.factories import create_user
from django_spire.knowledge.entry.tests.factories import create_test_entry
from django_spire.knowledge.entry.version.models import EntryVersion


def create_test_entry_version(**kwargs) -> EntryVersion:
    data = {
        'entry': kwargs.pop('entry', None) or create_test_entry(),
        'author': kwargs.pop('author', None) or create_user(username=random_string(k=10)),
        'is_deleted': False,
        'is_active': True
    }
    data.update(kwargs)
    return EntryVersion.objects.create(**data)



from __future__ import annotations

from dandy import Prompt

from django_spire.ai.context.intelligence.prompts.organization_prompts import (
    organization_info_prompt
)


def chat_system_prompt() -> Prompt:
    return (
        Prompt()
        .text(
            'You are a chat assistant, please use the organization information below for answering questions.'
        )
        .line_break()
        .prompt(organization_info_prompt())
        .line_break()
        .text('Make sure to follow the rules below')
        .list(
            [
                'Do not talk about being an AI at all to maintain the illusion of being a human assistant.',
            ]
        )
    )



from __future__ import annotations

from django import forms

class MFAForm(forms.Form):
    pass

    # def __init__(self, generated_mfa_code, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #
    #     self.helper = FormHelper(self)
    #     self.helper.include_media = False
    #     self.generated_mfa_code = generated_mfa_code
    #     self.fields['mfa_code'] = forms.CharField(max_length=6, label='Authentication Code')
    #
    #     self.helper.layout = Layout(
    #         Row(
    #             Column('mfa_code', css_class='form-group col-12'),
    #         ),
    #         HTML("{% include 'django_spire/contrib/form/form_submit_button.html' %}")
    #     )

    def clean(self) -> None:
        cleaned_data = super().clean()

        mfa_code = cleaned_data.get('mfa_code')

        if mfa_code != self.generated_mfa_code.code:
            self.add_error('mfa_code', 'Invalid MFA Code.')



from __future__ import annotations

from typing import ClassVar

from django import forms

from django_spire.help_desk.models import HelpDeskTicket


class HelpDeskTicketCreateForm(forms.ModelForm):
    class Meta:
        model = HelpDeskTicket
        exclude: ClassVar[list] = ['created_by', 'status']


class HelpDeskTicketUpdateForm(forms.ModelForm):
    class Meta:
        model = HelpDeskTicket
        exclude: ClassVar[list] = ['created_by']



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.email.tests.factories import create_test_email_notification
from django_spire.notification.models import Notification


class EmailNotificationModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.email_notification = create_test_email_notification()

    def test_str(self):
        expected = f'{self.email_notification.notification.title} - {self.email_notification.to_email_address}'
        assert str(self.email_notification) == expected

    def test_notification_relationship(self):
        assert self.email_notification.notification is not None
        assert isinstance(self.email_notification.notification, Notification)

    def test_default_template_id(self):
        assert self.email_notification.template_id == ''

    def test_default_context_data(self):
        assert self.email_notification.context_data == {}

    def test_default_cc(self):
        assert self.email_notification.cc == []

    def test_default_bcc(self):
        assert self.email_notification.bcc == []

    def test_to_email_address(self):
        assert self.email_notification.to_email_address == 'test@example.com'

    def test_with_cc(self):
        email_notification = create_test_email_notification(
            cc=['cc1@example.com', 'cc2@example.com']
        )
        assert email_notification.cc == ['cc1@example.com', 'cc2@example.com']

    def test_with_bcc(self):
        email_notification = create_test_email_notification(
            bcc=['bcc1@example.com', 'bcc2@example.com']
        )
        assert email_notification.bcc == ['bcc1@example.com', 'bcc2@example.com']

    def test_with_context_data(self):
        email_notification = create_test_email_notification(
            context_data={'key': 'value'}
        )
        assert email_notification.context_data == {'key': 'value'}



from __future__ import annotations

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.collection.tests.factories import create_test_collection
from django_spire.knowledge.entry.models import Entry
from django_spire.knowledge.entry.tests.factories import create_test_entry
from django_spire.knowledge.entry.version.choices import EntryVersionStatusChoices
from django_spire.knowledge.entry.version.tests.factories import create_test_entry_version


class EntryQuerySetTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.user = create_user(username='test_entry_queryset_user')
        self.collection = create_test_collection()
        self.entry = create_test_entry(collection=self.collection)
        self.entry_version = create_test_entry_version(entry=self.entry)
        self.entry.current_version = self.entry_version
        self.entry.save()

    def test_has_current_version(self):
        result = Entry.objects.has_current_version()
        assert self.entry in result

    def test_has_current_version_excludes_none(self):
        entry_without_version = create_test_entry(
            collection=self.collection,
            name='No Version'
        )
        result = Entry.objects.has_current_version()
        assert entry_without_version not in result

    def test_id_in(self):
        entry2 = create_test_entry(collection=self.collection, name='Entry 2')
        result = Entry.objects.id_in([self.entry.pk, entry2.pk])
        assert self.entry in result
        assert entry2 in result

    def test_user_has_access_published(self):
        self.entry_version.status = EntryVersionStatusChoices.PUBLISHED
        self.entry_version.save()

        result = Entry.objects.user_has_access(user=self.user)
        assert self.entry in result

    def test_user_has_access_draft_own(self):
        self.entry_version.status = EntryVersionStatusChoices.DRAFT
        self.entry_version.author = self.user
        self.entry_version.save()

        result = Entry.objects.user_has_access(user=self.user)
        assert self.entry in result

    def test_user_has_access_draft_other(self):
        other_user = create_user(username='other_user')
        self.entry_version.status = EntryVersionStatusChoices.DRAFT
        self.entry_version.author = other_user
        self.entry_version.save()

        result = Entry.objects.user_has_access(user=self.user)
        assert self.entry not in result

    def test_get_by_version_block_id(self):
        from django_spire.knowledge.entry.version.block.tests.factories import (
            create_test_version_block,
        )

        version_block = create_test_version_block(version=self.entry_version)
        result = Entry.objects.get_by_version_block_id(version_block_id=version_block.pk)
        assert result == self.entry



from __future__ import annotations

from django.utils.timezone import now, timedelta

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.app.tests.factories import create_test_app_notification
from django_spire.notification.models import Notification


class AppNotificationModelTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.app_notification = create_test_app_notification()

    def test_str(self):
        assert str(self.app_notification) == self.app_notification.notification.title

    def test_verbose_time_since_delivered_just_now(self):
        self.app_notification.notification.sent_datetime = now()
        self.app_notification.notification.save()

        assert self.app_notification.verbose_time_since_delivered == 'just now'

    def test_verbose_time_since_delivered_minutes(self):
        self.app_notification.notification.sent_datetime = now() - timedelta(minutes=5)
        self.app_notification.notification.save()

        assert '5 minutes ago' in self.app_notification.verbose_time_since_delivered

    def test_verbose_time_since_delivered_hours(self):
        self.app_notification.notification.sent_datetime = now() - timedelta(hours=3)
        self.app_notification.notification.save()

        assert '3 hours ago' in self.app_notification.verbose_time_since_delivered

    def test_verbose_time_since_delivered_days(self):
        self.app_notification.notification.sent_datetime = now() - timedelta(days=2)
        self.app_notification.notification.save()

        assert '2 days ago' in self.app_notification.verbose_time_since_delivered

    def test_verbose_time_since_delivered_singular_day(self):
        self.app_notification.notification.sent_datetime = now() - timedelta(days=1)
        self.app_notification.notification.save()

        assert '1 day ago' in self.app_notification.verbose_time_since_delivered

    def test_as_dict(self):
        result = self.app_notification.as_dict()

        assert result['id'] == self.app_notification.id
        assert result['title'] == self.app_notification.notification.title
        assert result['body'] == self.app_notification.notification.body
        assert result['url'] == self.app_notification.notification.url
        assert result['priority'] == self.app_notification.notification.priority
        assert 'time_since_delivered' in result
        assert 'context_data' in result

    def test_as_json(self):
        result = self.app_notification.as_json()
        assert isinstance(result, str)
        assert self.app_notification.notification.title in result

    def test_notification_relationship(self):
        assert self.app_notification.notification is not None
        assert isinstance(self.app_notification.notification, Notification)

    def test_default_template(self):
        assert self.app_notification.template == 'django_spire/notification/app/item/notification_item.html'

    def test_default_context_data(self):
        assert self.app_notification.context_data == {}



from __future__ import annotations

from django.urls import reverse

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.knowledge.collection.tests.factories import create_test_collection
from django_spire.knowledge.entry.tests.factories import create_test_entry


class EntryJsonUrlsTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.collection = create_test_collection()
        self.test_entry = create_test_entry(collection=self.collection)

    def test_reorder_view_url_path(self):
        response = self.client.post(
            reverse('django_spire:knowledge:entry:json:reorder'),
            data={
                'entry_id': self.test_entry.pk,
                'order': 0,
                'collection_id': self.collection.pk
            },
            content_type='application/json',
        )
        assert response.status_code == 200

    def test_reorder_view_invalid_entry(self):
        response = self.client.post(
            reverse('django_spire:knowledge:entry:json:reorder'),
            data={
                'entry_id': 99999,
                'order': 0,
                'collection_id': self.collection.pk
            },
            content_type='application/json',
        )
        assert response.status_code == 200
        assert response.json()['type'] == 'error'

    def test_reorder_view_missing_order(self):
        response = self.client.post(
            reverse('django_spire:knowledge:entry:json:reorder'),
            data={
                'entry_id': self.test_entry.pk,
                'collection_id': self.collection.pk
            },
            content_type='application/json',
        )
        assert response.status_code == 200
        assert response.json()['type'] == 'error'

    def test_reorder_view_invalid_collection(self):
        response = self.client.post(
            reverse('django_spire:knowledge:entry:json:reorder'),
            data={
                'entry_id': self.test_entry.pk,
                'order': 0,
                'collection_id': 99999
            },
            content_type='application/json',
        )
        assert response.status_code == 200
        assert response.json()['type'] == 'error'

    def test_update_files_view_url_path(self):
        response = self.client.post(
            reverse('django_spire:knowledge:entry:json:update_files'),
            content_type='application/json',
        )
        assert response.status_code == 200
        assert 'files_json' in response.json()



from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.notification.app.models import AppNotification
from django_spire.notification.choices import (
    NotificationStatusChoices,
    NotificationTypeChoices,
)
from django_spire.notification.exceptions import NotificationError
from django_spire.notification.models import Notification
from django_spire.notification.processors.notification import NotificationProcessor


class NotificationProcessorTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.user = create_user(username='test_processor_user')
        self.processor = NotificationProcessor()

    def _create_notification(self, **kwargs) -> Notification:
        data = {
            'user': self.user,
            'type': NotificationTypeChoices.APP,
            'title': 'Test Notification',
            'body': 'Test body',
            'status': NotificationStatusChoices.PENDING,
        }
        data.update(kwargs)
        notification = Notification.objects.create(**data)

        if data['type'] == NotificationTypeChoices.APP:
            AppNotification.objects.create(notification=notification)

        return notification

    @patch('django_spire.notification.processors.notification.AppNotificationProcessor')
    def test_process_app_notification(self, mock_processor_class: MagicMock):
        mock_processor = MagicMock()
        mock_processor_class.return_value = mock_processor

        notification = self._create_notification(type=NotificationTypeChoices.APP)

        self.processor.process(notification)

        mock_processor.process.assert_called_once_with(notification)

    @patch('django_spire.notification.processors.notification.EmailNotificationProcessor')
    def test_process_email_notification(self, mock_processor_class: MagicMock):
        mock_processor = MagicMock()
        mock_processor_class.return_value = mock_processor

        notification = self._create_notification(type=NotificationTypeChoices.EMAIL)

        self.processor.process(notification)

        mock_processor.process.assert_called_once_with(notification)

    @patch('django_spire.notification.processors.notification.SMSNotificationProcessor')
    def test_process_sms_notification(self, mock_processor_class: MagicMock):
        mock_processor = MagicMock()
        mock_processor_class.return_value = mock_processor

        notification = self._create_notification(type=NotificationTypeChoices.SMS)

        self.processor.process(notification)

        mock_processor.process.assert_called_once_with(notification)

    def test_process_unknown_type_raises_error(self):
        notification = self._create_notification()
        notification.type = 'unknown'

        with pytest.raises(NotificationError):
            self.processor.process(notification)

    def test_process_unknown_type_sets_status_failed(self):
        notification = self._create_notification()
        notification.type = 'unknown'

        with pytest.raises(NotificationError):
            self.processor.process(notification)

        notification.refresh_from_db()
        assert notification.status == NotificationStatusChoices.FAILED

    @patch('django_spire.notification.processors.notification.AppNotificationProcessor')
    def test_process_list_groups_by_type(self, mock_processor_class: MagicMock):
        mock_processor = MagicMock()
        mock_processor_class.return_value = mock_processor

        notifications = [
            self._create_notification(type=NotificationTypeChoices.APP),
            self._create_notification(type=NotificationTypeChoices.APP),
        ]

        self.processor.process_list(notifications)

        mock_processor.process_list.assert_called_once()
        call_args = mock_processor.process_list.call_args[0][0]
        assert len(call_args) == 2



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class FilterForm:
    """Playwright component for django_spire/filtering/form/base_session_filter_form.html"""

    def __init__(self, page: Page, form_selector: str = 'form[action*="filter"], form[action*="search"]') -> None:
        self.form_selector = form_selector
        self.page = page

    @property
    def clear_button(self) -> Locator:
        return self.form.locator('input[value="Clear"], button:has-text("Clear")')

    @property
    def filter_button(self) -> Locator:
        return self.form.locator('input[value="Filter"], button:has-text("Filter")')

    @property
    def form(self) -> Locator:
        return self.page.locator(self.form_selector).first

    @property
    def search_button(self) -> Locator:
        return self.form.locator('input[value="Search"], button:has-text("Search")')

    @property
    def search_input(self) -> Locator:
        return self.form.locator('input[name="search"], input[placeholder*="Search"]')

    def clear(self) -> None:
        self.clear_button.click()

    def fill_field(self, name: str, value: str) -> None:
        self.form.locator(f'[name="{name}"]').fill(value)

    def filter(self) -> None:
        self.filter_button.click()

    def get_field_value(self, name: str) -> str:
        return self.form.locator(f'[name="{name}"]').input_value()

    def search(self, query: str) -> None:
        self.search_input.fill(query)
        self.search_button.click()

    def select_option(self, name: str, value: str) -> None:
        self.form.locator(f'select[name="{name}"]').select_option(value)

    def submit(self) -> None:
        submit_btn = self.form.locator('button[type="submit"], input[type="submit"]').first
        submit_btn.click()



from __future__ import annotations

from unittest.mock import MagicMock

from django.core.paginator import Page
from django.test import RequestFactory, TestCase

from django_spire.contrib.pagination.pagination import paginate_list
from django_spire.contrib.pagination.templatetags.pagination_tags import (
    get_elided_page_range,
    pagination_url,
)


class TestPaginateList(TestCase):
    def setUp(self) -> None:
        self.object_list = list(range(100))

    def test_custom_page_number(self) -> None:
        result = paginate_list(self.object_list, page_number=2)

        assert result.number == 2

    def test_custom_per_page(self) -> None:
        result = paginate_list(self.object_list, per_page=10)

        assert len(result.object_list) == 10

    def test_default_page_number(self) -> None:
        result = paginate_list(self.object_list)

        assert result.number == 1

    def test_default_per_page(self) -> None:
        result = paginate_list(self.object_list)

        assert len(result.object_list) == 50

    def test_empty_list(self) -> None:
        result = paginate_list([])

        assert len(result.object_list) == 0

    def test_has_next_page(self) -> None:
        result = paginate_list(self.object_list, page_number=1)

        assert result.has_next() is True

    def test_has_previous_page(self) -> None:
        result = paginate_list(self.object_list, page_number=2)

        assert result.has_previous() is True

    def test_last_page_has_no_next(self) -> None:
        result = paginate_list(self.object_list, page_number=2)

        assert result.has_next() is False

    def test_page_number_exceeds_pages_returns_last_page(self) -> None:
        result = paginate_list(self.object_list, page_number=999)

        assert result.number == 2

    def test_returns_page_object(self) -> None:
        result = paginate_list(self.object_list)

        assert isinstance(result, Page)


class TestPaginationUrl(TestCase):
    def setUp(self) -> None:
        self.factory = RequestFactory()

    def _create_context(self, request) -> MagicMock:
        context = MagicMock()
        context.request = request
        return context

    def test_custom_page_name(self) -> None:
        request = self.factory.get('/')
        context = self._create_context(request)

        result = pagination_url(context, page_number=2, page_name='p')

        assert 'p=2' in result

    def test_default_page_name(self) -> None:
        request = self.factory.get('/')
        context = self._create_context(request)

        result = pagination_url(context, page_number=1)

        assert 'page=1' in result

    def test_multiple_query_params(self) -> None:
        request = self.factory.get('/', {'filter': 'active', 'sort': 'name'})
        context = self._create_context(request)

        result = pagination_url(context, page_number=3)

        assert 'page=3' in result
        assert 'filter=active' in result
        assert 'sort=name' in result

    def test_preserves_existing_query_params(self) -> None:
        request = self.factory.get('/', {'search': 'test'})
        context = self._create_context(request)

        result = pagination_url(context, page_number=2)

        assert 'search=test' in result
        assert 'page=2' in result

    def test_replaces_space_with_plus(self) -> None:
        request = self.factory.get('/', {'search': 'hello world'})
        context = self._create_context(request)

        result = pagination_url(context, page_number=1)

        assert 'search=hello+world' in result

    def test_returns_query_string(self) -> None:
        request = self.factory.get('/')
        context = self._create_context(request)

        result = pagination_url(context, page_number=1)

        assert result.startswith('?')


class TestGetElidedPageRange(TestCase):
    def test_custom_on_each_side(self) -> None:
        page_obj = MagicMock()
        page_obj.number = 5
        page_obj.paginator.get_elided_page_range.return_value = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        get_elided_page_range(page_obj, on_each_side=3)

        page_obj.paginator.get_elided_page_range.assert_called_once_with(
            number=5,
            on_each_side=3,
            on_ends=2
        )

    def test_custom_on_ends(self) -> None:
        page_obj = MagicMock()
        page_obj.number = 5
        page_obj.paginator.get_elided_page_range.return_value = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        get_elided_page_range(page_obj, on_ends=3)

        page_obj.paginator.get_elided_page_range.assert_called_once_with(
            number=5,
            on_each_side=2,
            on_ends=3
        )

    def test_default_parameters(self) -> None:
        page_obj = MagicMock()
        page_obj.number = 1
        page_obj.paginator.get_elided_page_range.return_value = [1, 2, 3]

        get_elided_page_range(page_obj)

        page_obj.paginator.get_elided_page_range.assert_called_once_with(
            number=1,
            on_each_side=2,
            on_ends=2
        )

    def test_returns_elided_page_range(self) -> None:
        page_obj = MagicMock()
        page_obj.number = 1
        expected_range = [1, 2, 3, '...', 10]
        page_obj.paginator.get_elided_page_range.return_value = expected_range

        result = get_elided_page_range(page_obj)

        assert result == expected_range



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import messages

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.forms import Form


def form_errors_as_list(form: Form) -> list[str]:
    form_errors = []

    for field_name, error_list in form.errors.items():
        for error in error_list.data:
            error_message = ''

            if field_name != '__all__':
                error_message += f'{field_name.title()}: '

            if hasattr(error, 'message_responses'):
                error_message += f'{" ".join(error.message_responses)}'

            elif hasattr(error, 'messages'):
                error_message += f'{" ".join(error.messages)}'

            else:
                message = 'Error message not found.'
                raise Exception(message)

            form_errors.append(error_message)

    return form_errors


def show_form_errors(request: WSGIRequest, *forms: Form) -> None:
    for form in forms:
        for error in form_errors_as_list(form):
            messages.error(request=request, message=error)



from __future__ import annotations

import json
import random
import string

from typing import TYPE_CHECKING

from django import template

if TYPE_CHECKING:
    from typing import Iterable

    from django.contrib.messages.storage.base import Message


register = template.Library()


@register.simple_tag()
def django_messages_to_json(messages: Iterable[Message]) -> str:
    return json.dumps([
        {
            'id': ''.join(random.choice(string.ascii_letters) for _ in range(8)),
            'message': message.message,
            'type': message.level_tag
        }
        for message in messages
    ])



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.service import BaseDjangoModelService
from django_spire.knowledge.entry.version.block.services.factory_service import EntryVersionBlockFactoryService

if TYPE_CHECKING:
    from django_spire.knowledge.entry.version.block.models import EntryVersionBlock


class EntryVersionBlockService(BaseDjangoModelService['EntryVersionBlock']):
    obj: EntryVersionBlock
    factory = EntryVersionBlockFactoryService()



from __future__ import annotations

from enum import StrEnum


class ProgressStatus(StrEnum):
    COMPLETE = 'complete'
    COMPLETING = 'completing'
    ERROR = 'error'
    PENDING = 'pending'
    RUNNING = 'running'



from __future__ import annotations

from django.test import TestCase

from django_spire.contrib.seeding.field.callable import CallableFieldSeeder


class TestCallableFieldSeeder(TestCase):
    def test_seeds_callable_fields(self) -> None:
        fields = {
            'timestamp': lambda: '2024-01-01 12:00:00',
            'status': lambda: 'active'
        }
        seeder = CallableFieldSeeder(fields=fields)

        result = seeder.seed(None, count=3)

        assert len(result) == 3

        for row in result:
            assert row['timestamp'] == '2024-01-01 12:00:00'
            assert row['status'] == 'active'



from __future__ import annotations

from dandy import BaseIntel, Bot, recorder_to_html_file

from django_spire.ai.decorators import log_ai_interaction_from_recorder
from django_spire.ai.models import AiInteraction, AiUsage
from django_spire.core.tests.test_cases import BaseTestCase


class HorseIntel(BaseIntel):
    breed: str
    color: str
    first_name: str
    has_cone_taped_to_head: bool


class AiDecoratorTestCase(BaseTestCase):
    def test_ai_interaction_decorator_creates_usage_record(self) -> None:
        initial_count = AiUsage.objects.count()

        @log_ai_interaction_from_recorder(self.super_user, 'horse')
        @recorder_to_html_file('horse')
        def generate_horse_intel(user_input: str) -> HorseIntel:
            bot = Bot()
            return bot.llm.prompt_to_intel(
                prompt=user_input,
                intel_class=HorseIntel,
            )

        horse_intel = generate_horse_intel('Make me a magical horse that grants wishes!')

        assert horse_intel.first_name != ''
        assert AiUsage.objects.count() >= initial_count

    def test_ai_interaction_decorator_creates_interaction_record(self) -> None:
        initial_count = AiInteraction.objects.count()

        @log_ai_interaction_from_recorder(self.super_user, 'test_actor')
        @recorder_to_html_file('test_interaction')
        def generate_horse_intel(user_input: str) -> HorseIntel:
            bot = Bot()
            return bot.llm.prompt_to_intel(
                prompt=user_input,
                intel_class=HorseIntel,
            )

        generate_horse_intel('Create a horse')

        assert AiInteraction.objects.count() > initial_count

    def test_ai_interaction_decorator_requires_user_or_actor(self) -> None:
        try:
            @log_ai_interaction_from_recorder()
            def dummy_func() -> None:
                pass
        except ValueError as e:
            assert 'user or actor must be provided' in str(e)
        else:
            assert False, 'Expected ValueError'

    def test_ai_interaction_decorator_with_actor_only(self) -> None:
        @log_ai_interaction_from_recorder(actor='test_actor_only')
        @recorder_to_html_file('test_actor_only')
        def generate_horse_intel(user_input: str) -> HorseIntel:
            bot = Bot()
            return bot.llm.prompt_to_intel(
                prompt=user_input,
                intel_class=HorseIntel,
            )

        horse_intel = generate_horse_intel('Create a horse')

        assert horse_intel is not None

        interaction = AiInteraction.objects.filter(actor='test_actor_only').first()
        assert interaction is not None
        assert interaction.actor == 'test_actor_only'

    def test_ai_interaction_decorator_records_module_and_callable(self) -> None:
        @log_ai_interaction_from_recorder(self.super_user, 'module_test')
        @recorder_to_html_file('module_test')
        def test_callable(user_input: str) -> HorseIntel:
            bot = Bot()
            return bot.llm.prompt_to_intel(
                prompt=user_input,
                intel_class=HorseIntel,
            )

        test_callable('Test input')

        interaction = AiInteraction.objects.filter(actor='module_test').first()
        assert interaction is not None
        assert 'test_callable' in interaction.callable_name


class AiUsageModelTestCase(BaseTestCase):
    def test_ai_usage_str(self) -> None:
        ai_usage = AiUsage.objects.create()

        assert 'ai usage' in str(ai_usage)

    def test_ai_usage_default_values(self) -> None:
        ai_usage = AiUsage.objects.create()

        assert ai_usage.event_count == 0
        assert ai_usage.token_usage == 0
        assert ai_usage.run_time_seconds == 0.0
        assert ai_usage.was_successful is True


class AiInteractionModelTestCase(BaseTestCase):
    def test_ai_interaction_str(self) -> None:
        ai_usage = AiUsage.objects.create()
        ai_interaction = AiInteraction.objects.create(
            ai_usage=ai_usage,
            actor='test_actor',
            module_name='test_module',
            callable_name='test_callable',
        )

        assert 'test_actor' in str(ai_interaction)
        assert 'interaction' in str(ai_interaction)

    def test_ai_interaction_saves_user_info(self) -> None:
        ai_usage = AiUsage.objects.create()
        ai_interaction = AiInteraction.objects.create(
            ai_usage=ai_usage,
            user=self.super_user,
            actor=None,
            module_name='test_module',
            callable_name='test_callable',
        )

        assert ai_interaction.user_email == self.super_user.email
        assert ai_interaction.user_first_name == self.super_user.first_name
        assert ai_interaction.user_last_name == self.super_user.last_name

    def test_ai_interaction_default_values(self) -> None:
        ai_usage = AiUsage.objects.create()
        ai_interaction = AiInteraction.objects.create(
            ai_usage=ai_usage,
            actor='test',
            module_name='test',
            callable_name='test',
        )

        assert ai_interaction.event_count == 0
        assert ai_interaction.token_usage == 0
        assert ai_interaction.run_time_seconds == 0.0
        assert ai_interaction.was_successful is True
        assert ai_interaction.exception is None
        assert ai_interaction.stack_trace is None



from __future__ import annotations

import json

from django.contrib.auth.models import User
from django.test import RequestFactory

from django_spire.core.shortcuts import (
    get_object_or_none,
    get_object_or_null_obj,
    model_object_from_app_label,
    process_request_body
)
from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.core.tests.test_models import DummyModel, DummyModelMixin


class ShortcutsTestCase(DummyModelMixin, BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.dummy = DummyModel.objects.create(name='test')
        self.user = User.objects.create_user(username='testuser', password='testpass')  # noqa: S106
        self.request_factory = RequestFactory()

    def test_get_object_or_none_found(self) -> None:
        obj = get_object_or_none(DummyModel, pk=self.dummy.pk)
        assert obj == self.dummy

    def test_get_object_or_none_not_found(self) -> None:
        obj = get_object_or_none(DummyModel, pk=9999)
        assert obj is None

    def test_get_object_or_null_obj_model_found(self) -> None:
        obj = get_object_or_null_obj(DummyModel, name='test')
        assert obj == self.dummy

    def test_get_object_or_null_obj_model_not_found(self) -> None:
        obj = get_object_or_null_obj(DummyModel, name='nonexistent')
        assert obj.pk is None
        assert obj.name == ''

    def test_get_object_or_null_obj_queryset_found(self) -> None:
        obj = get_object_or_null_obj(DummyModel.objects.all(), name='test')
        assert obj == self.dummy

    def test_get_object_or_null_obj_queryset_not_found(self) -> None:
        obj = get_object_or_null_obj(DummyModel.objects.all(), name='nonexistent')
        assert obj.pk is None
        assert obj.name == ''

    def test_model_object_from_app_label_contenttype_not_found(self) -> None:
        obj = model_object_from_app_label('nonexistent', 'nonexistent', 1)
        assert obj is None

    def test_model_object_from_app_label_found(self) -> None:
        obj = model_object_from_app_label('auth', 'user', self.user.pk)
        assert obj == self.user

    def test_model_object_from_app_label_not_found(self) -> None:
        obj = model_object_from_app_label('auth', 'user', 9999)
        assert obj is None

    def test_process_request_body(self) -> None:
        data = {'data': {'key': 'value'}}

        request = self.request_factory.post(
            '/dummy-url/',
            data=json.dumps(data),
            content_type='application/json'
        )

        result = process_request_body(request)
        assert result == data['data']



from __future__ import annotations

from django_spire.history.querysets import HistoryQuerySet


class OrganizationQuerySet(HistoryQuerySet):
    def get_only_or_none(self):
        try:
            return self.earliest('id')
        except self.model.DoesNotExist:
            return None


class PeopleQuerySet(HistoryQuerySet):
    pass



from __future__ import annotations

from typing import Callable, TYPE_CHECKING

from dandy import Decoder

from django_spire.conf import settings
from django_spire.core.utils import get_callable_from_module_string_and_validate_arguments

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def generate_intent_decoder(
    request: WSGIRequest,
    default_callable: Callable | None = None
) -> Decoder:
    intent_dict = {}

    if hasattr(settings, 'DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS'):
        for intent_config in settings.DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS.values():
            required_permission = intent_config.get('REQUIRED_PERMISSION')

            if required_permission and not request.user.has_perm(required_permission):
                continue

            intent_description = intent_config.get('INTENT_DESCRIPTION', '')
            chat_router_path = intent_config.get('CHAT_ROUTER')

            if chat_router_path:
                try:
                    router_class = get_callable_from_module_string_and_validate_arguments(
                        chat_router_path,
                        []
                    )

                    router_instance = router_class()

                    intent_dict[intent_description] = router_instance.workflow
                except ImportError:
                    pass

    if default_callable is not None:
        intent_dict['None of the above choices match the user\'s intent'] = default_callable

    return Decoder(
        mapping_keys_description='Intent of the User\'s Request',
        mapping=intent_dict
    )



from __future__ import annotations

import threading

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable

    from django_spire.contrib.progress.session import ProgressSession


class ParallelTask:
    def __init__(self, session: ProgressSession, task_id: str, future: Any) -> None:
        self._future = future
        self._session = session
        self._task_id = task_id

        self._session.start(task_id)

        thread = threading.Thread(
            target=self._wait_for_completion,
            daemon=True,
        )
        thread.start()

    def _wait_for_completion(self) -> None:
        try:
            _ = self._future.result
            self._session.complete(self._task_id)
        except Exception:
            self._session.error(self._task_id)

    @property
    def result(self) -> Any:
        return self._future.result


class SequentialTask:
    def __init__(
        self,
        session: ProgressSession,
        task_id: str,
        func: Callable,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self._exception: Exception | None = None
        self._result: Any = None
        self._session = session
        self._task_id = task_id

        self._session.start(task_id)

        try:
            self._result = func(*args, **kwargs)
            self._session.complete(task_id)
        except Exception as e:
            self._exception = e
            self._session.error(task_id)

    @property
    def result(self) -> Any:
        if self._exception is not None:
            raise self._exception

        return self._result



from __future__ import annotations

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.choices import (
    HelpDeskTicketPriorityChoices,
    HelpDeskTicketPurposeChoices,
    HelpDeskTicketStatusChoices,
)
from django_spire.help_desk.forms import (
    HelpDeskTicketCreateForm,
    HelpDeskTicketUpdateForm,
)
from django_spire.help_desk.tests.factories import create_test_helpdesk_ticket


class HelpDeskTicketCreateFormTests(BaseTestCase):
    def test_valid_form(self):
        form_data = {
            'purpose': HelpDeskTicketPurposeChoices.APP,
            'priority': HelpDeskTicketPriorityChoices.LOW,
            'description': 'Test description',
        }
        form = HelpDeskTicketCreateForm(data=form_data)
        assert form.is_valid()

    def test_invalid_form_missing_purpose(self):
        form_data = {
            'priority': HelpDeskTicketPriorityChoices.LOW,
            'description': 'Test description',
        }
        form = HelpDeskTicketCreateForm(data=form_data)
        assert not form.is_valid()
        assert 'purpose' in form.errors

    def test_invalid_form_missing_priority(self):
        form_data = {
            'purpose': HelpDeskTicketPurposeChoices.APP,
            'description': 'Test description',
        }
        form = HelpDeskTicketCreateForm(data=form_data)
        assert not form.is_valid()
        assert 'priority' in form.errors

    def test_invalid_form_missing_description(self):
        form_data = {
            'purpose': HelpDeskTicketPurposeChoices.APP,
            'priority': HelpDeskTicketPriorityChoices.LOW,
        }
        form = HelpDeskTicketCreateForm(data=form_data)
        assert not form.is_valid()
        assert 'description' in form.errors

    def test_excludes_created_by(self):
        assert 'created_by' not in HelpDeskTicketCreateForm().fields

    def test_excludes_status(self):
        assert 'status' not in HelpDeskTicketCreateForm().fields


class HelpDeskTicketUpdateFormTests(BaseTestCase):
    def setUp(self):
        super().setUp()
        self.ticket = create_test_helpdesk_ticket()

    def test_valid_form(self):
        form_data = {
            'purpose': HelpDeskTicketPurposeChoices.COMPANY,
            'priority': HelpDeskTicketPriorityChoices.HIGH,
            'status': HelpDeskTicketStatusChoices.INPROGRESS,
            'description': 'Updated description',
        }
        form = HelpDeskTicketUpdateForm(data=form_data, instance=self.ticket)
        assert form.is_valid()

    def test_includes_status(self):
        assert 'status' in HelpDeskTicketUpdateForm().fields

    def test_excludes_created_by(self):
        assert 'created_by' not in HelpDeskTicketUpdateForm().fields

    def test_invalid_form_missing_status(self):
        form_data = {
            'purpose': HelpDeskTicketPurposeChoices.APP,
            'priority': HelpDeskTicketPriorityChoices.LOW,
            'description': 'Test description',
        }
        form = HelpDeskTicketUpdateForm(data=form_data, instance=self.ticket)
        assert not form.is_valid()
        assert 'status' in form.errors



from __future__ import annotations

from django.contrib.admin.sites import AdminSite
from django.test import RequestFactory

from django_spire.core.tests.test_cases import BaseTestCase
from django_spire.help_desk.admin import HelpDeskTicketAdmin
from django_spire.help_desk.models import HelpDeskTicket
from django_spire.help_desk.tests.factories import create_test_helpdesk_ticket


class HelpDeskTicketAdminTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.site = AdminSite()
        self.admin = HelpDeskTicketAdmin(HelpDeskTicket, self.site)
        self.factory = RequestFactory()
        self.ticket = create_test_helpdesk_ticket()

    def test_list_display(self):
        expected = ('pk', 'purpose', 'priority', 'status', 'created_by', 'created_datetime')
        assert self.admin.list_display == expected

    def test_list_filter(self):
        expected = ('priority', 'purpose', 'status', 'is_active', 'is_deleted')
        assert self.admin.list_filter == expected

    def test_ordering(self):
        assert self.admin.ordering == ('-created_datetime',)

    def test_raw_id_fields(self):
        assert self.admin.raw_id_fields == ('created_by',)

    def test_readonly_fields(self):
        expected = ('created_by', 'created_datetime', 'is_active', 'is_deleted')
        assert self.admin.readonly_fields == expected

    def test_search_fields(self):
        expected = ('description', 'created_by__username', 'created_by__email')
        assert self.admin.search_fields == expected



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from dandy.recorder import recorder_to_html_file
from dandy import SqliteCache, generate_cache_key

from django_spire.contrib.seeding.field.override import FieldOverride
from django_spire.contrib.seeding.model.config import FieldsConfig

if TYPE_CHECKING:
    from typing import Any


class classproperty:
    def __init__(self, fget: Any) -> None:
        self.fget = fget

    def __get__(self, instance: Any, owner: Any) -> Any:
        return self.fget(owner)


class BaseModelSeeder(ABC):
    model_class = None

    fields = None
    field_config_class = FieldsConfig

    default_to = 'llm'

    cache_seed = True
    cache_name = 'model_seeder'
    cache_limit = 1000

    override_class = FieldOverride

    _field_seeders = []
    _field_config: FieldsConfig = None

    @classproperty
    def override(cls) -> FieldOverride:
        return cls.override_class(seeder_class=cls)

    @classmethod
    def get_field_config(cls) -> FieldsConfig:
        if cls._field_config is None:
            if cls.model_class is None:
                message = 'model_class must be defined before using seeder.'
                raise ValueError(message)

            raw_fields = cls.__dict__.get('fields', {})

            cls._field_config = cls.field_config_class(
                raw_fields=raw_fields,
                field_names=cls.field_names(),
                default_to=cls.default_to,
                model_class=cls.model_class,
            )

        return cls._field_config

    @classmethod
    def resolved_fields(cls) -> dict:
        return cls.get_field_config().fields

    @classmethod
    def clear_cache(cls) -> None:
        SqliteCache.clear(cache_name=cls.cache_name)

    @classmethod
    @abstractmethod
    def field_names(cls) -> list[str]:
        pass

    @classmethod
    @recorder_to_html_file('model_seeder')
    def seed_data(
        cls,
        count: int = 1,
        fields: dict | None = None,
    ) -> list[dict]:
        field_config = (
            cls.get_field_config().override(fields)
            if fields
            else cls.get_field_config()
        )

        if cls.cache_seed:
            cache_key = generate_cache_key(
                cls.seed_data, count=count, fields=field_config.fields
            )

            cache = SqliteCache(cache_name=cls.cache_name, limit=cls.cache_limit)
            formatted_seed_data = cache.get(cache_key)

            if formatted_seed_data:
                return formatted_seed_data

        seed_data = []

        for seeder_cls in cls._field_seeders:
            seeder = seeder_cls(field_config.fields, field_config.default_to)

            if len(seeder.seeder_fields) > 0:
                seed_data.append(seeder.seed(cls, count))

        formatted_seed_data = [
            {} for _ in range(max(len(sublist) for sublist in seed_data))
        ]
        for sublist in seed_data:
            for i, d in enumerate(sublist):
                formatted_seed_data[i].update(d)

        if cls.cache_seed:
            cache.set(cache_key, formatted_seed_data)

        return formatted_seed_data

    @classmethod
    def seed(
        cls,
        count: int = 1,
        fields: dict | None = None,
    ) -> list:
        return [
            cls.model_class(**seed_data) for seed_data in cls.seed_data(count, fields)
        ]



from __future__ import annotations

from django.db.models.functions import Coalesce
from django.db.models import JSONField

from django_spire.contrib.ordering.querysets import OrderingQuerySetMixin
from django_spire.history.querysets import HistoryQuerySet
from django.db.models import Value


class EntryVersionBlockQuerySet(HistoryQuerySet, OrderingQuerySetMixin):
    def format_for_editor(self):
        coalesce_json_field = lambda field_name: Coalesce(
            field_name,
            Value({}, output_field=JSONField())
        )

        return (
            self.annotate(
                data=coalesce_json_field('_block_data'),
                tunes=coalesce_json_field('_tunes_data'),
            )
            .order_by('order')
            .values('id', 'type', 'data', 'tunes')
        )


from __future__ import annotations

from django.contrib import admin

from django_spire.core.tag.models import Tag


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    list_display = ('name',)
    ordering = ('name',)
    search_fields = ('name',)



from __future__ import annotations

from django.db import models


class HelpDeskTicketPurposeChoices(models.TextChoices):
    APP = ('app', 'App')
    COMPANY = ('comp', 'Company')


class HelpDeskTicketStatusChoices(models.TextChoices):
    READY = ('read', 'Ready')
    INPROGRESS = ('prog', 'In Progress')
    DONE = ('done', 'Done')


class HelpDeskTicketPriorityChoices(models.TextChoices):
    LOW = ('low', 'Low')
    MEDIUM = ('med', 'Medium')
    HIGH = ('high', 'High')
    URGENT = ('urge', 'Urgent')



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.auth.permissions.consts import (
    PERMISSIONS_LEVEL_CHOICES,
    VALID_PERMISSION_LEVELS
)

if TYPE_CHECKING:
    from django.contrib.auth.models import Group, User


def set_group_users(group: Group, user_list: list[User]) -> None:
    group.user_set.set(user_list)


def codename_to_perm_level(codename: str) -> int:
    return perm_level_to_int(codename.split('_')[0])


def codename_list_to_perm_level(codenames_list: list[str]) -> int:
    perm_level = 0

    for codename in codenames_list:
        temp_perm_level = perm_level_to_int(codename.split('_')[0])

        if temp_perm_level > perm_level:
            perm_level = temp_perm_level

    return perm_level


def perm_level_to_int(perm_level: VALID_PERMISSION_LEVELS | str) -> int:
    if isinstance(perm_level, int) and 4 >= perm_level >= 0:
        return perm_level

    if isinstance(perm_level, str):
        for perm in PERMISSIONS_LEVEL_CHOICES:
            if perm_level.lower() == perm[1].lower():
                return perm[0]

    return 0


def perm_level_to_string(perm_level: VALID_PERMISSION_LEVELS | str) -> str:
    perm_level = perm_level_to_int(perm_level)

    for perm in PERMISSIONS_LEVEL_CHOICES:
        if perm_level == perm[0]:
            return perm[1].capitalize()

    return 'None'


def perm_level_to_django_permission(
        perm_level: VALID_PERMISSION_LEVELS,
        app_label: str,
        model_name: str
) -> str:
    perm_level = dict(PERMISSIONS_LEVEL_CHOICES)[perm_level_to_int(perm_level)]
    return f'{app_label}.{perm_level.lower()}_{model_name}'


def has_app_permission(
    user: User,
    app_label: str,
    model_name: str,
    action: str
) -> bool:
    return user.has_perm(f'{app_label}.{action}_{model_name}')


def has_app_permission_or_404(
    user: User,
    app_label: str,
    model_name: str,
    action: str
) -> bool:
    if not has_app_permission(user, app_label, model_name, action):
        raise PermissionError

    return True



from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import HttpResponse, HttpResponseForbidden
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from twilio.twiml.messaging_response import MessagingResponse

from django_spire.ai.sms.decorators import twilio_auth_required
from django_spire.ai.sms.intelligence.workflows.sms_conversation_workflow import sms_conversation_workflow
from django_spire.ai.sms.models import SmsConversation

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@csrf_exempt
@require_POST
@twilio_auth_required
def webhook_view(request: WSGIRequest):
    from_number = request.POST.get('From', '')

    if len(from_number) < 11:
        return HttpResponseForbidden()

    body = request.POST.get('Body', '')
    message_sid = request.POST.get('MessageSid', '')

    conversation, _ = SmsConversation.objects.get_or_create(
        phone_number=from_number
    )

    message = conversation.add_message(
        body=body,
        is_inbound=True,
        twilio_sid=message_sid,
    )

    sms_intel = sms_conversation_workflow(
        request=request,
        user_input=body,
        message_history=conversation.generate_message_history(),
        actor=from_number,
    )

    twiml_response = MessagingResponse()
    twiml_response.message(sms_intel.body)

    conversation.add_message(
        body=sms_intel.body,
        is_inbound=False,
        twilio_sid=message_sid,
        is_processed=True
    )

    message.is_processed = True
    message.save()

    return HttpResponse(twiml_response)



from __future__ import annotations

from dandy.llm.request.message import MessageHistory, RoleLiteralStr
from django.contrib.auth.models import User
from django.db import models
from django.utils.timezone import now

from django_spire.ai.sms.querysets import SmsConversationQuerySet, SmsMessageQuerySet
from django_spire.history.mixins import HistoryModelMixin


class SmsConversation(HistoryModelMixin):
    user = models.ForeignKey(
        User,
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        related_name='sms_conversations',
        related_query_name='sms_conversation'
    )

    phone_number = models.CharField(max_length=20)
    last_message_datetime = models.DateTimeField(default=now, editable=False)

    objects = SmsConversationQuerySet.as_manager()

    def __str__(self):
        return f"SMS Conversation with {self.phone_number}"

    def add_message(
        self,
        body: str,
        is_inbound: bool,
        twilio_sid: str,
        is_processed: bool = False
    ):
        message = self.messages.create(
            body=body,
            is_inbound=is_inbound,
            twilio_sid=twilio_sid,
            is_processed=is_processed,
        )

        self.last_message_datetime = now()
        self.save()

        return message

    def generate_message_history(
        self,
        message_count: int = 20,
        exclude_last_message: bool = True
    ) -> MessageHistory:
        message_history = MessageHistory()

        messages = self.messages.newest_by_count(message_count)

        if exclude_last_message:
            messages = messages[1:]

        messages = list(reversed(messages))

        for message in messages:
            message_history.add_message(
                role=message.role,
                content=message.body
            )

        return message_history

    @property
    def is_empty(self) -> bool:
        return self.messages.count() == 0

    class Meta:
        db_table = 'django_spire_ai_sms_conversation'
        verbose_name = 'SMS Conversation'
        verbose_name_plural = 'SMS Conversations'
        ordering = ('-last_message_datetime',)


class SmsMessage(HistoryModelMixin):
    conversation = models.ForeignKey(
        SmsConversation,
        on_delete=models.CASCADE,
        related_name='messages',
        related_query_name='message'
    )

    body = models.TextField()

    is_inbound = models.BooleanField(default=False)

    twilio_sid = models.CharField(max_length=64, blank=True, null=True)

    is_processed = models.BooleanField(default=False)

    objects = SmsMessageQuerySet.as_manager()

    def __str__(self):
        if len(self.body) < 64:
            return f"{self.direction}: {self.body}"

        return f"{self.direction}: {self.body[:64]}..."

    @property
    def direction(self) -> str:
        return "Inbound" if self.is_inbound else "Outbound"

    @property
    def is_outbound(self) -> bool:
        return not self.is_inbound

    @property
    def role(self) -> RoleLiteralStr:
        if self.is_inbound:
            return 'user'

        if self.is_outbound:
            return 'assistant'

        return 'system'

    class Meta:
        db_table = 'django_spire_ai_sms_message'
        verbose_name = 'SMS Message'
        verbose_name_plural = 'SMS Messages'
        ordering = ('-created_datetime',)



from __future__ import annotations

from typing import TYPE_CHECKING

from django.conf import settings
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.urls import reverse

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@login_required()
def login_redirect_view(request: WSGIRequest, **kwargs):
    if not request.user.is_authenticated:
        return HttpResponseRedirect(reverse('django_spire:auth:admin:login'))

    # MFA implementation if required
    # profile = request.user.profile
    #
    # if profile.requires_mfa():
    #     return HttpResponseRedirect(reverse('django_spire:auth:mfa:redirect:notification'))

    return HttpResponseRedirect(reverse(settings.LOGIN_REDIRECT_SUCCESS_URL))


@login_required()
def logout_redirect_view(request: WSGIRequest, **kwargs):
    logout(request)
    return HttpResponseRedirect(reverse('django_spire:auth:admin:login'))



from __future__ import annotations

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.user.tests.factories import create_user
from django_spire.core.tests.test_cases import BaseTestCase


class AuthGroupModelTestCase(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.group = AuthGroup.objects.create(name='Test Group')

    def test_str_representation(self) -> None:
        assert str(self.group) == 'Test Group'

    def test_str_representation_with_special_characters(self) -> None:
        group = AuthGroup.objects.create(name='Test & Group <Special>')
        assert str(group) == 'Test & Group <Special>'

    def test_base_breadcrumb_returns_breadcrumbs(self) -> None:
        crumbs = AuthGroup.base_breadcrumb()
        assert crumbs is not None

    def test_breadcrumbs_with_pk_returns_breadcrumbs(self) -> None:
        crumbs = self.group.breadcrumbs()
        assert crumbs is not None

    def test_breadcrumbs_without_pk_returns_breadcrumbs(self) -> None:
        unsaved_group = AuthGroup(name='Unsaved Group')
        crumbs = unsaved_group.breadcrumbs()
        assert crumbs is not None

    def test_meta_proxy(self) -> None:
        assert AuthGroup._meta.proxy

    def test_meta_verbose_name(self) -> None:
        assert AuthGroup._meta.verbose_name == 'Auth Group'
        assert AuthGroup._meta.verbose_name_plural == 'Auth Groups'

    def test_group_creation(self) -> None:
        group = AuthGroup.objects.create(name='New Test Group')
        assert group.pk is not None
        assert AuthGroup.objects.filter(pk=group.pk).exists()

    def test_group_update(self) -> None:
        self.group.name = 'Updated Group Name'
        self.group.save()
        self.group.refresh_from_db()
        assert self.group.name == 'Updated Group Name'

    def test_group_deletion(self) -> None:
        group_pk = self.group.pk
        self.group.delete()
        assert not AuthGroup.objects.filter(pk=group_pk).exists()

    def test_group_user_relationship(self) -> None:
        user = create_user(username='groupuser')
        self.group.user_set.add(user)
        assert user in self.group.user_set.all()
        assert self.group in user.groups.all()

    def test_group_permissions_relationship(self) -> None:
        assert hasattr(self.group, 'permissions')

    def test_group_name_max_length(self) -> None:
        max_length = AuthGroup._meta.get_field('name').max_length
        assert max_length == 150

    def test_multiple_groups_creation(self) -> None:
        groups = [
            AuthGroup.objects.create(name=f'Group {i}')
            for i in range(5)
        ]
        assert len(groups) == 5
        assert AuthGroup.objects.count() >= 5

    def test_group_queryset_ordering(self) -> None:
        AuthGroup.objects.create(name='Zebra Group')
        AuthGroup.objects.create(name='Alpha Group')
        groups = AuthGroup.objects.all().order_by('name')
        names = list(groups.values_list('name', flat=True))
        assert names == sorted(names)

    def test_group_add_multiple_users(self) -> None:
        user1 = create_user(username='user1')
        user2 = create_user(username='user2')
        self.group.user_set.add(user1, user2)
        assert self.group.user_set.count() == 2

    def test_group_remove_user(self) -> None:
        user = create_user(username='removeuser')
        self.group.user_set.add(user)
        self.group.user_set.remove(user)
        assert user not in self.group.user_set.all()

    def test_group_clear_users(self) -> None:
        user1 = create_user(username='clearuser1')
        user2 = create_user(username='clearuser2')
        self.group.user_set.add(user1, user2)
        self.group.user_set.clear()
        assert self.group.user_set.count() == 0

    def test_group_add_permission(self) -> None:
        content_type = ContentType.objects.get_for_model(AuthGroup)
        permission = Permission.objects.filter(content_type=content_type).first()
        if permission:
            self.group.permissions.add(permission)
            assert permission in self.group.permissions.all()

    def test_group_remove_permission(self) -> None:
        content_type = ContentType.objects.get_for_model(AuthGroup)
        permission = Permission.objects.filter(content_type=content_type).first()
        if permission:
            self.group.permissions.add(permission)
            self.group.permissions.remove(permission)
            assert permission not in self.group.permissions.all()

    def test_group_with_unicode_name(self) -> None:
        group = AuthGroup.objects.create(name='Tëst Grøup 日本語')
        assert group.name == 'Tëst Grøup 日本語'
        assert str(group) == 'Tëst Grøup 日本語'

    def test_group_name_with_max_length(self) -> None:
        long_name = 'A' * 150
        group = AuthGroup.objects.create(name=long_name)
        assert group.name == long_name

    def test_breadcrumbs_contains_group_name(self) -> None:
        crumbs = self.group.breadcrumbs()
        names = [c['name'] for c in crumbs]
        assert 'Test Group' in names

    def test_base_breadcrumb_contains_groups(self) -> None:
        crumbs = AuthGroup.base_breadcrumb()
        names = [c['name'] for c in crumbs]
        assert 'Groups' in names

    def test_group_user_set_query(self) -> None:
        user = create_user(username='queryuser')
        self.group.user_set.add(user)
        queried_users = self.group.user_set.filter(username='queryuser')
        assert queried_users.count() == 1

    def test_group_inheritance_from_django_group(self) -> None:
        from django.contrib.auth.models import Group
        assert issubclass(AuthGroup, Group)

    def test_group_activity_mixin(self) -> None:
        assert hasattr(self.group, 'add_activity')

    def test_group_empty_name_constraint(self) -> None:
        group = AuthGroup.objects.create(name='')
        assert group.pk is not None

    def test_group_whitespace_name(self) -> None:
        group = AuthGroup.objects.create(name='   ')
        assert group.name == '   '

    def test_group_special_characters_name(self) -> None:
        special_name = '!@#$%^&*()_+-=[]{}|;:,.<>?'
        group = AuthGroup.objects.create(name=special_name)
        assert group.name == special_name

    def test_group_refresh_from_db(self) -> None:
        original_name = self.group.name
        AuthGroup.objects.filter(pk=self.group.pk).update(name='Changed Name')
        self.group.refresh_from_db()
        assert self.group.name == 'Changed Name'
        assert self.group.name != original_name

    def test_group_get_or_create(self) -> None:
        group, created = AuthGroup.objects.get_or_create(name='Get Or Create Group')
        assert created
        group2, created2 = AuthGroup.objects.get_or_create(name='Get Or Create Group')
        assert not created2
        assert group.pk == group2.pk

    def test_group_bulk_create(self) -> None:
        groups = AuthGroup.objects.bulk_create([
            AuthGroup(name='Bulk 1'),
            AuthGroup(name='Bulk 2'),
            AuthGroup(name='Bulk 3'),
        ])
        assert len(groups) == 3

    def test_group_exists(self) -> None:
        assert AuthGroup.objects.filter(name='Test Group').exists()
        assert not AuthGroup.objects.filter(name='Nonexistent Group').exists()



from __future__ import annotations

from unittest.mock import patch

from django.test import TestCase

from django_spire.contrib.help.templatetags.help import help_button


class TestHelpButton(TestCase):
    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_generates_unique_help_id(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html')

        calls = mock_render.call_args_list
        help_id_1 = calls[0][0][1]['help_id']

        mock_render.reset_mock()

        help_button('test_template.html')

        calls = mock_render.call_args_list
        help_id_2 = calls[0][0][1]['help_id']

        assert help_id_1 != help_id_2

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_help_id_starts_with_help_prefix(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html')

        calls = mock_render.call_args_list
        context = calls[0][0][1]

        assert context['help_id'].startswith('help-')

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_help_title_defaults_to_none(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html')

        calls = mock_render.call_args_list
        button_context = calls[0][0][1]

        assert button_context['help_title'] is None

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_passes_help_title_to_button_template(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html', help_title='Test Title')

        calls = mock_render.call_args_list
        button_context = calls[0][0][1]

        assert button_context['help_title'] == 'Test Title'

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_passes_help_title_to_modal_template(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html', help_title='Test Title')

        calls = mock_render.call_args_list

        # Order: button (0), help_content (1), modal (2)
        modal_context = calls[2][0][1]

        assert modal_context['help_title'] == 'Test Title'

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_renders_button_and_modal_templates(self, mock_render: patch) -> None:
        mock_render.return_value = ''

        help_button('test_template.html')

        assert mock_render.call_count == 3

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_returns_concatenated_button_and_modal(self, mock_render: patch) -> None:
        # Order of calls: button, help_content, modal
        mock_render.side_effect = ['<button>', '<content>', '<modal>']

        result = help_button('test_template.html')

        assert '<button>' in result
        assert '<modal>' in result

    @patch('django_spire.contrib.help.templatetags.help.render_to_string')
    def test_renders_help_content_template(self, mock_render: patch) -> None:
        mock_render.return_value = '<content>'

        help_button('custom_help.html')

        template_names = [call[0][0] for call in mock_render.call_args_list]
        assert 'custom_help.html' in template_names



from __future__ import annotations

import logging

from abc import ABC
from typing import Generic, TypeVar

from django.db import transaction
from django.db.models import Model

from django_spire.contrib.constructor.django_model_constructor import BaseDjangoModelConstructor
from django_spire.contrib.service.exceptions import ServiceError


log = logging.getLogger(__name__)

TypeDjangoModel_co = TypeVar('TypeDjangoModel_co', bound=Model, covariant=True)


class BaseDjangoModelService(
    BaseDjangoModelConstructor[TypeDjangoModel_co],
    ABC,
    Generic[TypeDjangoModel_co]
):
    def _get_concrete_fields(self) -> dict:
        return {
            field.name: field
            for field in self.obj._meta.get_fields()
            if field.concrete and not field.many_to_many and not field.one_to_many
        }

    def _get_touched_fields(self, concrete_fields: dict, **field_data: dict) -> list[str]:
        foreign_key_id_aliases = {f"{name}_id" for name, field in concrete_fields.items() if field.many_to_one}
        allowed = set(concrete_fields) | foreign_key_id_aliases

        touched_fields: list[str] = []

        for field, value in field_data.items():
            if field not in allowed:
                log.warning(f'Field {field!r} is not valid for {self.obj.__class__.__name__}')
                continue

            model_field = concrete_fields.get(field.removesuffix("_id"), None)

            if model_field and (getattr(model_field, 'auto_created', False) or not model_field.editable):
                continue

            setattr(self.obj, field, value)

            touched_fields.append(field.removesuffix('_id'))

        return touched_fields

    def validate_model_obj(self, **field_data: dict) -> list[str]:
        concrete_fields = self._get_concrete_fields()
        touched_fields = self._get_touched_fields(concrete_fields, **field_data)

        exclude = [field for field in concrete_fields if field not in touched_fields]
        self.obj.full_clean(exclude=exclude)

        return touched_fields

    @transaction.atomic
    def save_model_obj(self, **field_data: dict) -> tuple[Model, bool]:
        new_model_obj_was_created = False

        if not field_data:
            message = f'Field data is required to save on {self.obj.__class__.__name__}'
            raise ServiceError(message)

        touched_fields = self.validate_model_obj(**field_data)

        if self.model_obj_is_new:
            new_model_obj_was_created = True
            self.obj.save()

        elif touched_fields:
            self.obj.save(update_fields=touched_fields)

        else:
            message = f'{self.obj.__class__.__name__} is not a new object or there was no touched fields to update.'
            log.warning(message)

        return self.obj, new_model_obj_was_created
